#include <time.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <signal.h>
#include "aggregation.h"
#include "qerror.h"
#include "random.h"

#define MAIN_GROUP_SIZE 10000
//#define AGGREGATION_NOTREE

char * MAIN_FILE = "abilene.gra";

int DISPLAY = -1;

// We consider that a new group "group" is arriving in the network. A shortest path tree or a minimum spanning tree "tree_g" is computed for this group considering the choosed options (spt or mst).

int NUMBER_OF_CANDIDATES = 1;

int CREATE_NEW_TREE_MST = 0; // allows to create minimum spanning trees (MST) as new trees, instead of shortest path trees

float STA2_BOUND = 0.0;
int STA2_MINREQ = 1;
float AQOSM_BOUND = 0.0;
// main [spt | mst] (number 100000) (display 10000) (graph nsfnet.gra) (bound 0.0) 

// spt allows STA to create shortest path tree for new trees while mst allows STA to create Minimum Spanning tree
// number 100000 determines the total number of groups which will arrive in the network. In the example STA will aggregate 100000 groups

// display 10000 determines the frequences of the display. In the example, informations are displayed every 10000 groups

// graph nsfnet.gra determines the network on which STA will make aggregations. You can use also eurorings.gra.

// bound 0.0 defines the bandwidth threshold tb. In the example, tb=0.0 and no bandwidth is allowed to be wasted. 

// example : main  mst number 100000 display 10000 graph nsfnet.gra bound 0.2 

//When only typing main the default parameters are : Shortest path tree algorithm (by chosing an arbitrary source), 100000 groups, display every 10000 groups on the graph eurorings.gra with a bandwidth threshold of 0.2 :main  spt number 10000 display 10000 graph eurorings.gra bound 0.2 

/***********************************************************************************/
/****************************** main function **************************************/
/***********************************************************************************/


void aggregationMain(int argc,char ** argv) {
	struct aggregation * a1;
	struct aggregation * a2;
	struct aggregation * a3;
	struct ugraph * g;
	struct bandwidth * b;
	int res1;
	int res2;
	int res3;
	struct nodeWeightedModel * model;
	struct intlist * group;
	int groupSize;
	int groupRequirement;
	int totalNumberOfGroups = MAIN_GROUP_SIZE;
	int totalNumberOfTrees = MAIN_GROUP_SIZE;
	int i = 1;
	int numberEdgeRouters;
	int a1low = 0,a2low = 0,a3low = 0,a1medium = 0,a2medium = 0,a3medium = 0,a1high = 0,a2high = 0,a3high = 0;
	int low =0,medium =0,high=0;
	//int a1lowcost = 0,a1mediumcost = 0,a1highcost = 0;
	//int a2highcost = 0,a2mediumcost = 0,a2lowcost = 0;
	//int a3lowcost = 0,a3mediumcost = 0,a3highcost = 0;
	/* parameters */
	while (i<argc) {
		if (strcmp(argv[i],"spt")==0) {
			CREATE_NEW_TREE_MST = 0;
		}
		if (strcmp(argv[i],"mst")==0) {
			CREATE_NEW_TREE_MST = 1;
		}
		if (strcmp(argv[i],"number")==0) {
			totalNumberOfGroups = atoi(argv[i+1]); 
			totalNumberOfTrees = atoi(argv[i+1]); 
			i = i+1;
		}
		if (strcmp(argv[i],"graph")==0) {
			MAIN_FILE = argv[i+1];
			i = i+1;
		}
		if (strcmp(argv[i],"display")==0) {
			DISPLAY = atoi(argv[i+1]);
			i = i+1;
		}
		if (strcmp(argv[i],"stabound")==0) {
			STA2_BOUND = atof(argv[i+1]);
			i = i+1;
		}
		if (strcmp(argv[i],"aqosmbound")==0) {
			AQOSM_BOUND = atof(argv[i+1]);
			i = i+1;
		}
		if (strcmp(argv[i],"randseed")==0) {
			srandom(atoi(argv[i+1]));
			i = i+1;
		}
		if (strcmp(argv[i],"candidates")==0) {
			NUMBER_OF_CANDIDATES = atoi(argv[i+1]);
			i = i+1;
		}
		if (strcmp(argv[i],"stareq")==0) {
			STA2_MINREQ = atoi(argv[i+1]);
			i = i+1;
		}
		i = i+1;
	}
	if (DISPLAY==-1) {
	  if(totalNumberOfGroups==10000){
	    DISPLAY = 10000;
	  }
	  else{
	    DISPLAY = totalNumberOfGroups/10;
	    if (DISPLAY>10000) {
	      DISPLAY = 10000;
	    }
	  }
	}
	/* initialization */
	if (strcmp(MAIN_FILE,"random.gra")==0) {
		g = ugraphCreateDoar(50,0.25,0.2);
		numberEdgeRouters = AGGREGATION_EDGEROUTERS_FOR_RANDOM;
		if (numberEdgeRouters>g->nodeNumber) {
			numberEdgeRouters = g->nodeNumber;
		}
	}
	else {
		g = ugraphCreateFromFile(MAIN_FILE);
		numberEdgeRouters = g->nodeNumber;
	}
	b = bandwidthCreate(g);
	a1 = aggregationCreate(g,totalNumberOfGroups,totalNumberOfTrees,numberEdgeRouters,b);
	a2 = aggregationCreate(g,totalNumberOfGroups,totalNumberOfTrees,numberEdgeRouters,b);
	a3 = aggregationCreate(g,totalNumberOfGroups,totalNumberOfTrees,numberEdgeRouters,b);
	model = nodeWeightedModelCreate(g->nodeNumber);
	/* aggregation */
	for (i=0;i<totalNumberOfGroups;i++) {

		/*a new random group "group" is created*/
		groupSize = randomGet(a1->numberEdgeRouters-2)+2;
		//group = randomList(groupSize,g->nodeNumber);
		group = nodeWeightedModelNewGroup(model,groupSize);

		/*The group is added to the structure aggregation "a" with a transformation of group with bitmap*/
		groupRequirement = randomGet(100);
		if (groupRequirement<50) {
		  groupRequirement = 1;
		}
		else {
		  if (groupRequirement<80) {
		    groupRequirement = 10;
		  }
		  else {
		    groupRequirement = 100;
		  }
		}

		res1 = aggregationAddGroupAQOSM(a1,aggregationIntlistToBitmap(group),groupRequirement);
		res2 = aggregationAddGroupNoAggregation(a2,aggregationIntlistToBitmap(group),groupRequirement);
		res3 = aggregationAddGroupSTA2(a3,aggregationIntlistToBitmap(group),groupRequirement);
		switch (groupRequirement) {
		case 1:
		  low ++;
		  if (res1) {
		    a1low++;
		    //a1lowcost=a1lowcost+utreeTotalCost(a1->trees[a1->groupToTree[i]]);
		  }
		  if (res2) {
		    a2low++;
		    //a2lowcost=a2lowcost+utreeTotalCost(a2->trees[a2->groupToTree[i]]);
		  }
		  if (res3) {
		    a3low++;
		    //a3lowcost=a3lowcost+utreeTotalCost(a3->trees[a3->groupToTree[i]]);
		  }
		  break;
		case 10:
		  medium++;
		  if (res1) {
		    a1medium++;
		    //a1mediumcost=a1mediumcost+utreeTotalCost(a1->trees[a1->groupToTree[i]]);
		  }
		  if (res2) {
		    a2medium++;
		    //a2mediumcost=a2mediumcost+utreeTotalCost(a2->trees[a2->groupToTree[i]]);
		  }
		  if (res3) {
		    a3medium++;
		    //a3mediumcost=a3mediumcost+utreeTotalCost(a3->trees[a3->groupToTree[i]]);
		  }
		  break;
		case 100:
		  high++;
		  if (res1) {
		    a1high++;
		    //a1highcost=a1highcost+utreeTotalCost(a1->trees[a1->groupToTree[i]]);
		  }
		  if (res2) {
		    a2high++;
		    //a2highcost=a2highcost+utreeTotalCost(a2->trees[a2->groupToTree[i]]);
		  }
		  if (res3) {
		    a3high++;
		    //a3highcost=a3highcost+utreeTotalCost(a3->trees[a3->groupToTree[i]]);
		  }
		  break;
		}
		
		if ((i+1)%DISPLAY==0) {
		  printf("QOS : %d %d %d %d %.2f %.2f %.2f",i+1,a1low,a1medium,a1high,(100.0*a1low)/low,(100.0*a1medium)/medium,(100.0*a1high)/high);
		  printf(" %d %d %lld %.2f %.2f\n",a1->numberOfAcceptedGroups,a1->numberOfTrees,a1->numberOfComparisons,(100.0*a1->numberOfTrees)/a1->numberOfAcceptedGroups,(100.0*aggregationGetTotalBandwidth(a1))/a1->numberOfAcceptedGroups);
		  printf("NO- : %d %d %d %d %.2f %.2f %.2f",i+1,a2low,a2medium,a2high,(100.0*a2low)/low,(100.0*a2medium)/medium,(100.0*a2high)/high);
		  printf(" %d %d %lld %.2f %.2f\n",a2->numberOfAcceptedGroups,a2->numberOfTrees,a2->numberOfComparisons,(100.0*a2->numberOfTrees)/a2->numberOfAcceptedGroups,(100.0*aggregationGetTotalBandwidth(a2))/a2->numberOfAcceptedGroups);
		  printf("STC : %d %d %d %d %.2f %.2f %.2f",i+1,a3low,a3medium,a3high,(100.0*a3low)/low,(100.0*a3medium)/medium,(100.0*a3high)/high);
		  printf(" %d %d %lld %.2f %.2f\n",a3->numberOfAcceptedGroups,a3->numberOfTrees,a3->numberOfComparisons,(100.0*a3->numberOfTrees)/a3->numberOfAcceptedGroups,(100.0*aggregationGetTotalBandwidth(a3))/a3->numberOfAcceptedGroups);
		}
		intlistFree(group);
	}
	/*printf("AQOSM\n");
	bandwidthDisplay(a1->bandwidth); printf("\n");aggregationDisplayStress(a1);
		printf("\n\nNo\n");	bandwidthDisplay(a2->bandwidth);printf("\n");aggregationDisplayStress(a2);
	printf("\n\nSTA\n");	bandwidthDisplay(a3->bandwidth);printf("\n");aggregationDisplayStress(a3);
	
	*/
	nodeWeightedModelFree(model);
}

/*************************************************************************************************/
/******************* Creation and Destruction ****************************************************/
/*************************************************************************************************/




/* creates a new aggregation structure for at most "maxNumberOfGroups" groups and at most "maxNumberOfTrees" trees */
struct aggregation * aggregationCreate(struct ugraph * graph,int maxNumberOfGroups,int maxNumberOfTrees,int numberEdgeRouters,struct bandwidth * bandwidth) {
	struct aggregation * a;
	int i;
	int maxTreeCost;
	if ((graph==NULL) || (maxNumberOfGroups<0) || (maxNumberOfTrees<0) || (numberEdgeRouters<0)) {
		QERROR("parameters");
	}
	if (numberEdgeRouters>8*sizeof(AGGREGATION_BITMAP)) {
		QERROR("AGGREGATION_BITMAP is not sufficient for the graph");
	}
	QERROR_MALLOC(a,1,struct aggregation);
	a->graph = graph;
	a->change =0;
	a->bandwidth = bandwidthCopy(bandwidth);
	a->numberEdgeRouters = numberEdgeRouters;
	a->maxNumberOfGroups = maxNumberOfGroups;
	a->numberOfGroups = 0;
	QERROR_MALLOC(a->groupRequirements,maxNumberOfGroups,int);
	QERROR_MALLOC(a->nativeTreeCost,maxNumberOfGroups,int);
	a->numberOfAcceptedGroups = 0;
	QERROR_MALLOC(a->groups,maxNumberOfGroups,AGGREGATION_BITMAP);
	a->maxNumberOfTrees = maxNumberOfTrees;
	a->numberOfTrees = 0;
#ifdef AGGREGATION_NOTREE
	QERROR_MALLOC(a->treecost,maxNumberOfTrees,int);
	QERROR_MALLOC(a->treenumberofspannednodes,maxNumberOfTrees,int);
	QERROR_MALLOC(a->treenumberofedges,maxNumberOfTrees,int);
#else
	QERROR_MALLOC(a->trees,maxNumberOfTrees,struct utree *);
#endif
	QERROR_MALLOC(a->treebitmap,maxNumberOfTrees,AGGREGATION_BITMAP);
	QERROR_MALLOC(a->groupToTree,maxNumberOfGroups,int);
	QERROR_MALLOC(a->numberOfAggregatedGroups,maxNumberOfTrees,int);
	a->numberOfComparisons = 0;
	a->maxNumberOfComparisons = 0;
	maxTreeCost = a->graph->nodeNumber*ugraphGetMaxEdgeCost(a->graph);
	QERROR_MALLOC(a->costset,maxTreeCost,struct intlist *);
	for (i=0;i<maxTreeCost;i++) {
	  a->costset[i] = intlistCreate();
	}
	return a;
}

/* frees "a" */
void aggregationFree(struct aggregation * a) {
	int i;
	if (a==NULL) {
	  QERROR("parameters");
	}
	bandwidthFree(a->bandwidth);
	free(a->groups);
	free(a->groupRequirements);
	free(a->nativeTreeCost);
#ifdef AGGREGATION_NOTREE
	free(a->treecost);
	free(a->treenumberofspannednodes);
	free(a->treenumberofedges);
#else
	for (i=0;i<a->numberOfTrees;i++) {
	  utreeFree(a->trees[i]);
	}
	free(a->trees);
#endif
	free(a->treebitmap);
	free(a->groupToTree);
	free(a->numberOfAggregatedGroups);
	for (i=0;i<a->graph->edgeNumber;i++) {
	  intlistFree(a->costset[i]);
	}
	free(a->costset);
	free(a);
}


/*************************************************************************************************/
/*******************      Display functions   ****************************************************/
/*************************************************************************************************/


/* displays "a": the total time to make all the aggregations, the percentage of aggregations, the number of forwarding entries in the routing table, the size of the forawarding table, and finally the average bandwidth used per group*/
void aggregationDisplay(struct aggregation * a) {
	if (a==NULL) {
		QERROR("parameters");
	}
	printf("\n");
	printf("For %d groups: \n",a->numberOfAcceptedGroups);
	printf("  %.2f acceptance ratio\n",(1.0*a->numberOfAcceptedGroups)/a->numberOfGroups);
	printf("  %d trees and aggregation ratio: %.2f\n",a->numberOfTrees,1.0-(1.0*a->numberOfTrees)/a->numberOfAcceptedGroups);
	printf("  forwarding table: %.2f entries per router and an average size of %.2f per router\n",(1.0*aggregationGetNumberOfForwardingEntries(a))/(a->graph->nodeNumber),(1.0*aggregationGetSizeOfForwardingTable(a))/a->graph->nodeNumber);
	printf("  average bandwidth used per group: %.2f\n",(1.0*aggregationGetTotalBandwidth(a))/a->numberOfAcceptedGroups);
	printf("  mean number of evaluated trees per group (even for unaccepted groups): %.2f\n",(1.0*a->numberOfComparisons)/a->numberOfGroups);
	printf("  maximum number of evaluated trees of a group: %d\n",a->maxNumberOfComparisons);
}

/* display the size of "costset" sets: the number of trees of the same cost  */
void aggregationGetTreesPerCost(struct aggregation * a) {
	int i;
	int maxEdgeCost;
	if (a==NULL) {
		QERROR("parameters");
	}
	if (a->costset!=NULL) {
		maxEdgeCost = ugraphGetMaxEdgeCost(a->graph);
		for (i=0;i<a->graph->nodeNumber*maxEdgeCost;i++) {
			if (intlistSize(a->costset[i])!=0) {
				printf("%d %d\n",i,intlistSize(a->costset[i]));
			}
		}
	}
}

float aggregationGetGroupRequirements(struct aggregation * a){
	int i;
	float requirements = 0;
	for(i=0;i<a->numberOfGroups;i++){
		if (a->groups[i]!=0) {
		  requirements = requirements + a->groupRequirements[i];	
		}
	}  
	return requirements;
}

/*returns the bandwidth used per group if no aggregation (i.e. the minimum bandwidth needed for all the groups as in traditional IP multicast)*/
int aggregationGetNativeBandwidth(struct aggregation * a) {
	int i;
	int costTrees = 0;
	
	for(i=0;i<a->numberOfGroups;i++){
		if (a->groups[i]!=0) {
			
		  costTrees = costTrees + a->nativeTreeCost[i];
		}
	}  
	return (costTrees);
}

/*returns the total number of members in the network*/
int aggregationGetTotalNumberOfMembers(struct aggregation * a) {
	int i;
	int members = 0;
	struct intlist * group;
	for(i=0;i<a->numberOfGroups;i++){
		if (a->groups[i]!=0) {
			group = aggregationBitmapToIntlist(a->groups[i]);
			members = members + intlistSize(group);
			
		}
	}  
	return (members);
}

/* returns the total bandwidth used for all the groups*/
int aggregationGetTotalBandwidth(struct aggregation * a) {
	int bandwidth = 0;
	int i;
	if (a==NULL) {
		QERROR("parameters");
	}
	for (i=0;i<a->numberOfGroups;i++) {
		if (a->groups[i]!=0) {
#ifdef AGGREGATION_NOTREE
			bandwidth = bandwidth+a->treecost[a->groupToTree[i]];
#else
			bandwidth = bandwidth+utreeTotalCost(a->trees[a->groupToTree[i]]);
#endif
		}
	}
	return bandwidth;
}

/* returns the total number of forwarding states: how many entries are created in the routing table for all the routing tables, the number of spanned nodes by all the trees*/
int aggregationGetNumberOfForwardingEntries(struct aggregation * a) {
	int entries = 0;
	int i;
	if (a==NULL) {
		QERROR("parameters");
	}
	for (i=0;i<a->numberOfTrees;i++) {
#ifdef AGGREGATION_NOTREE
		entries = entries+a->treenumberofspannednodes[i];
#else
		entries = entries+utreeNumberOfSpannedNodes(a->trees[i]);
#endif
	}
	return entries;
}

/* returns the total size of forwarding table, a forwarding entry has a size corresponding to the cost of a tree.  */
int aggregationGetSizeOfForwardingTable(struct aggregation * a) {
	int size = 0;
	int i;
	if (a==NULL) {
		QERROR("parameters");
	}
	for (i=0;i<a->numberOfTrees;i++) {
#ifdef AGGREGATION_NOTREE
		size = size+a->treenumberofedges[i];
#else
		size = size+utreeNumberOfEdges(a->trees[i]);
#endif
	}
	return size;
}



void aggregationDisplayStress(struct aggregation * a) {	
  
 int saturated = 0;
  int totalUse = 0;
  int i,j,edge;
  edge = 0;
  for (i=0;i<a->graph->nodeNumber;i++) {
    for (j=0;j<=i;j++) {
      if (ugraphIsEdge(a->graph,i,j)) {
	if (a->bandwidth->residual[edge]==0) {
	  saturated = saturated +1;
	}
	totalUse = totalUse +((a->bandwidth->capacity[edge]) - (a->bandwidth->residual[edge]));
	printf("arete %d-%d res %d cap %d use %f  \n",i,j,a->bandwidth->residual[edge],a->bandwidth->capacity[edge],(100.0*(a->bandwidth->capacity[edge] - a->bandwidth->residual[edge]))/a->bandwidth->capacity[edge]);
	//printf("totalUse %d \n",totalUse); 
	edge = edge+1;
      }
     
    }
  }
	
}
/* displays link usage */
#ifndef AGGREGATION_NOTREE
void aggregationLinkUsage(struct aggregation * a) {
	int group;
	
	int saturated = 0;
	int i,j;
	int edge;
	struct utree * t;
	int * linkUsage;
	if (a==NULL) {
		QERROR("parameters");
	}
	QERROR_MALLOC(linkUsage,a->graph->edgeNumber,int);
	for (edge=0;edge<a->graph->edgeNumber;edge++) {
		linkUsage[edge] = 0;
	}
	for (group=0;group<a->numberOfGroups;group++) {
		if (a->groups[i]==0) {
			t = a->trees[a->groupToTree[group]];
			edge = 0;
			for (i=0;i<a->graph->nodeNumber;i++) {
				for (j=0;j<i;j++) {
					if (ugraphIsEdge(a->graph,i,j)) {
						if (ugraphIsEdge(t->tree,i,j)) {
							linkUsage[edge] = linkUsage[edge]+1;
						}
						edge = edge+1;
					}
				}
			}
		}
	}
	//printf("links = ");
	edge = 0;
	for (i=0;i<a->graph->nodeNumber;i++) {
		for (j=0;j<i;j++) {
			if (ugraphIsEdge(a->graph,i,j)) {
				//    printf("%d ",linkUsage[edge]);
				if(linkUsage[edge] == a->bandwidth->capacity[edge]){
					saturated = saturated + 1;
				}
				edge = edge+1;
			}
		}
	}
	
}
#endif

/*************************************************************************************************/
/*******************   STA functions : Aggregation ***********************************************/
/*************************************************************************************************/


/* adds a new tree covering "group" to "a" using minimum spanning tree or shortest path tree algorithm*/
void aggregationAddNewTree(struct aggregation * a,AGGREGATION_BITMAP group) {
	struct utree * tree_g;
#ifndef AGGREGATION_NOAGG
	int source;
	struct intlist * g;
#endif
	if (a==NULL) {
		QERROR("parameters");
	}
	if (a->numberOfTrees+1>a->maxNumberOfTrees) {
		QERROR("too many trees");
	}
	g = aggregationBitmapToIntlist(group);
	source = g->head->value;
#ifndef AGGREGATION_NOAGG
	//g = aggregationBitmapToIntlist(group);
	//source = g->head->value;
	if (CREATE_NEW_TREE_MST==1) {
#ifdef AGGREGATION_NOTREE
		tree_g = utreeCreateMinimumSpanningTree(a->graph,source,g);
		a->treecost[a->numberOfTrees] = utreeTotalCost(tree_g);
#else
		tree_g = utreeCreateMinimumSpanningTree(a->graph,source,g);
		a->trees[a->numberOfTrees] = tree_g;
#endif
	}
	else {
#ifdef AGGREGATION_NOTREE
		tree_g = utreeCreateShortestPathTree(a->graph,source,g);
		a->treecost[a->numberOfTrees] = utreeTotalCost(tree_g);
#else
		tree_g = utreeCreateShortestPathTree(a->graph,source,g);
		a->trees[a->numberOfTrees] = tree_g;
#endif
	}
	intlistFree(g);
#ifdef AGGREGATION_NOTREE
	a->treenumberofedges[a->numberOfTrees] = utreeNumberOfEdges(tree_g);
	a->treenumberofspannednodes[a->numberOfTrees] = utreeNumberOfSpannedNodes(tree_g);
	a->treebitmap[a->numberOfTrees] = aggregationTreeBitmap(tree_g,a->numberEdgeRouters);
	utreeFree(tree_g);
#else
	a->treebitmap[a->numberOfTrees] = aggregationTreeBitmap(a->trees[a->numberOfTrees],a->numberEdgeRouters);
#endif
#endif
	a->groupToTree[a->numberOfGroups]= a->numberOfTrees;
	a->numberOfAggregatedGroups[a->numberOfTrees] = 1;
	a->numberOfTrees = a->numberOfTrees+1;
}

/* aggregate a group (the current group is "a->numberOfGroups") to a tree "tree" in "a" */
void aggregationAggregateToTree(struct aggregation * a,int tree) {
	if ((a==NULL) || (tree<0) ||(tree>=a->numberOfTrees)) {
		QERROR("parameters");
	}
	a->groupToTree[a->numberOfGroups] = tree;
	a->numberOfAggregatedGroups[tree] = a->numberOfAggregatedGroups[tree]+1;
}

/* adds "group" to "a" using STA and then decide whether or not to aggregate it to a tree*/
int aggregationAddGroupSTA1(struct aggregation * a,AGGREGATION_BITMAP group,int groupRequirement) {
	int available = 1;
	int i,source,found;
	int min_tree = -1;
	float min = AGGREGATION_INFINITY;
	struct utree * tree_g;
	struct utree * min_tree_candidate;
	float value;
	struct intlist_element * t;
	struct intlist * g;
	int maxComparisons = 0;
	int * candidates;
	int maxEdgeCost;
	QERROR_MALLOC(candidates,NUMBER_OF_CANDIDATES,int);
	for (i=0;i<NUMBER_OF_CANDIDATES;i++) {
		candidates[i] = -1;
	}
	a->groups[a->numberOfGroups] = group;
	a->groupRequirements[a->numberOfGroups]=groupRequirement;
	g = aggregationBitmapToIntlist(group);
	//intlistDisplay(g);	a->groupRequirements[a->numberOfGroups]=groupRequirement;
	source = g->head->value;
	//printf("source %d",source);
	if (CREATE_NEW_TREE_MST==1) {
		tree_g = utreeCreateMinimumSpanningTree(a->graph,source,g);
	}
	else {	//intlistDisplay(g);	printf("source %d",source);
		tree_g = utreeCreateShortestPathTree(a->graph,source,g);
	}
	//utreeDisplay(tree_g);
	//printf("\n\n\n\n\n\n");
	a->nativeTreeCost[a->numberOfGroups]=utreeTotalCost(tree_g);
	found = 0;
	i = (float)utreeTotalCost(tree_g);
	maxEdgeCost = ugraphGetMaxEdgeCost(a->graph);
	while ((i<maxEdgeCost*a->graph->nodeNumber) && (i<=utreeTotalCost(tree_g)*(1+STA2_BOUND)) && (found<NUMBER_OF_CANDIDATES)) {
	  t = a->costset[i]->head;
		while ((t!=NULL) && (found<NUMBER_OF_CANDIDATES)) {
			a->numberOfComparisons = a->numberOfComparisons+1;
			maxComparisons = maxComparisons+1;
			if (((~group)|(a->treebitmap[t->value]))==(AGGREGATION_BITMAP)-1) {
				min = value;
				candidates[found] = t->value;
				found = found+1;
			}
			t = t->next;
		}
		if (found<NUMBER_OF_CANDIDATES) {
			i = i+1;
		}
	}
	//printf("min %.2f\n",min);
	if (maxComparisons>a->maxNumberOfComparisons) {
		a->maxNumberOfComparisons = maxComparisons;
	}
	if (min==AGGREGATION_INFINITY) {
		if (bandwidthAddGroup(a->bandwidth,g,tree_g,groupRequirement)) {
		  //printf("bandwidthAddgroup OK\n");
			intlistAddToBack(a->costset[utreeTotalCost(tree_g)],a->numberOfTrees);
			aggregationAddNewTree(a,group);
			//printf("STA new tree of cost %d\n",utreeTotalCost(tree_g)); 
			//utreeDisplay(tree_g);
			//printf("\n");
			//utreeDisplay(tree_g);
			//printf("STA: tree_g\n");
			//printf("\n");
		   
		}
		else {
			available = 0;
			a->groups[a->numberOfGroups] = 0;
			//printf("STA tree_g not good\n");
			//utreeDisplay(tree_g); printf("\n");
		}
	}
	else {
		min_tree = aggregationCandidate(candidates);
#ifdef AGGREGATION_NOTREE
		QERROR("TODO");
#else
		min_tree_candidate = a->trees[min_tree];
		//printf("mintree STA: %d\n",min_tree);
#endif
		
		if (bandwidthAddGroup(a->bandwidth,g,min_tree_candidate,groupRequirement)) {
			aggregationAggregateToTree(a,min_tree);	
			//printf("STA:tree found of cost %d for the group with %d members\n",utreeTotalCost(a->trees[min_tree]),intlistSize(g));
			//utreeDisplay(a->trees[min_tree]);
			//printf("\n");
			//printf("ok pour STA aggregation\n");
		}
		else {
		  //printf("no agg\n");
		  available = 0;
		  a->groups[a->numberOfGroups] = 0;
		  //printf("STAmin_tree not good\n");
		  //utreeDisplay(a->trees[min_tree]);
		  //printf("\n");
		}
	}
	a->numberOfGroups++;
	if (available) {
		a->numberOfAcceptedGroups = a->numberOfAcceptedGroups+1;
	}
	else{
	  //printf("STA: group of size %d refused with tree of cost %d\n",intlistSize(g),utreeTotalCost(tree_g));  
	 
	}
	utreeFree(tree_g);
	intlistFree(g);
	return available;
}

/* adds "group" to "a" using STA and then decide whether or not to aggregate it to a tree*/
int aggregationAddGroupSTA2(struct aggregation * a,AGGREGATION_BITMAP group,int groupRequirement) {
	int available = 1;
	int i,source,found;
	int min_tree = -1;
	float min = AGGREGATION_INFINITY;
	struct utree * tree_g;
	struct utree * min_tree_candidate;
	float value;
	struct intlist_element * t;
	struct intlist * g;
	int maxComparisons = 0;
	int * candidates;
	int maxEdgeCost,minedgecost;
	struct ugraph * graphcopy;
	int k,l,edge;
	int cost_original = 0;
	QERROR_MALLOC(candidates,NUMBER_OF_CANDIDATES,int);
	for (i=0;i<NUMBER_OF_CANDIDATES;i++) {
		candidates[i] = -1;
	}
	a->groups[a->numberOfGroups] = group;
	a->groupRequirements[a->numberOfGroups]=groupRequirement;
	g = aggregationBitmapToIntlist(group);
	source = g->head->value;
	

	//	if (CREATE_NEW_TREE_MST==1) {
	// tree_g = utreeCreateMinimumSpanningTree(a->graph,source,g);
	  //}
	  //else {
	  //tree_g = utreeCreateShortestPathTree(a->graph,source,g);
	  //}

	cost_original = intlistSize(g)-1;



	minedgecost = 120000;
	do {
	  graphcopy = ugraphCreateFromFile(MAIN_FILE);
	  edge = 0;
	  for (k=0;k<a->graph->nodeNumber;k++) {
	    for (l=0;l<=k;l++) {
	      if (ugraphIsEdge(a->graph,k,l)) {
		if (a->bandwidth->residual[edge]<minedgecost) {
		  //if ((a->bandwidth->residual[edge]<ourThreshold) && (a->bandwidth->residual[edge]<groupRequirement)) {
		  ugraphRemoveEdges(graphcopy,k,l);
                   }
		edge = edge+1;
	      }
	    }
	  }
	  if (CREATE_NEW_TREE_MST==1) {
	    tree_g = utreeCreateMinimumSpanningTree(graphcopy,source,g);
	  }
	  else {
	    tree_g = utreeCreateShortestPathTree(graphcopy,source,g);
	  }
	  ugraphFree(graphcopy);
	  minedgecost = 9*minedgecost/10;
	} while ((tree_g==NULL) && (minedgecost>=groupRequirement)); 
	

if(tree_g == NULL){
		available = 0;
		a->groups[a->numberOfGroups] = 0;
	}
	else{
		a->nativeTreeCost[a->numberOfGroups]=utreeTotalCost(tree_g);
		found = 0;
		i = cost_original;
		maxEdgeCost = ugraphGetMaxEdgeCost(a->graph);
		//while ((i<maxEdgeCost*a->graph->nodeNumber) && (i<=utreeTotalCost(tree_g)*(1+1/(1+log(groupRequirement)))) && (found<NUMBER_OF_CANDIDATES)) {
		while ((i<maxEdgeCost*a->graph->nodeNumber) && (i<=utreeTotalCost(tree_g)) && (found<NUMBER_OF_CANDIDATES)) {
			//while ((i<maxEdgeCost*a->graph->nodeNumber) && (i<=utreeTotalCost(tree_g)) && (found<NUMBER_OF_CANDIDATES)) {
			t = a->costset[i]->head;
			while ((t!=NULL) && (found<NUMBER_OF_CANDIDATES)) {
				a->numberOfComparisons = a->numberOfComparisons+1;
				maxComparisons = maxComparisons+1;
				if (((~group)|(a->treebitmap[t->value]))==(AGGREGATION_BITMAP)-1) {
					if (bandwidthCheckGroup(a->bandwidth,g,a->trees[t->value],groupRequirement)) { // STA -> STA2
						min = value;
						candidates[found] = t->value;
						found = found+1;
					}
				}
				t = t->next;
			}
			if (found<NUMBER_OF_CANDIDATES) {
				i = i+1;
			}
		}
		//printf("min %.2f\n",min);
		if (maxComparisons>a->maxNumberOfComparisons) {
			a->maxNumberOfComparisons = maxComparisons;
		}
		if (min==AGGREGATION_INFINITY) {
			if (bandwidthAddGroup(a->bandwidth,g,tree_g,groupRequirement)) {
				intlistAddToFront(a->costset[utreeTotalCost(tree_g)],a->numberOfTrees); // (improvement ?)
				aggregationAddTree(a,group,tree_g);
			}
			else {
				available = 0;
				a->groups[a->numberOfGroups] = 0;
			}
		}
		else {
			min_tree = aggregationCandidate(candidates);
#ifdef AGGREGATION_NOTREE
			QERROR("TODO");
#else
			min_tree_candidate = a->trees[min_tree];
#endif
			if (bandwidthAddGroup(a->bandwidth,g,min_tree_candidate,groupRequirement)) {
				aggregationAggregateToTree(a,min_tree);	
			}
			else {
				available = 0;
				a->groups[a->numberOfGroups] = 0;
			}
		}
		//if (tree_g!=NULL) {
		//	utreeFree(tree_g);
		//}
	}
	a->numberOfGroups++;
	if (available) {
		a->numberOfAcceptedGroups = a->numberOfAcceptedGroups+1;
	}
	intlistFree(g);
	return available;
}

/* adds "group" to "a" using STA and then decide whether or not to aggregate it to a tree*/
int aggregationAddGroupSTA3(struct aggregation * a,AGGREGATION_BITMAP group,int groupRequirement) {
	int available = 1;
	int i,source,found;
	int edge,k,l;
	struct ugraph * graphcopy;
	int min_tree = -1;
	float min = AGGREGATION_INFINITY;
	struct utree * tree_g;
	struct utree * min_tree_candidate;
	float value;
	struct intlist_element * t;
	struct intlist * g;
	int maxComparisons = 0;
	int * candidates;
	int maxEdgeCost;
	QERROR_MALLOC(candidates,NUMBER_OF_CANDIDATES,int);
	for (i=0;i<NUMBER_OF_CANDIDATES;i++) {
		candidates[i] = -1;
	}
	a->groups[a->numberOfGroups] = group;
	a->groupRequirements[a->numberOfGroups]=groupRequirement;
	// STA -> STA3
	g = aggregationBitmapToIntlist(group);
	source = g->head->value;
	graphcopy = ugraphCreateFromFile(MAIN_FILE);
	edge = 0;
	for (k=0;k<a->graph->nodeNumber;k++) {
		for (l=0;l<=k;l++) {
			if (ugraphIsEdge(a->graph,k,l)) {
				if (a->bandwidth->residual[edge]<groupRequirement) {
					ugraphRemoveEdges(graphcopy,k,l);
				}
				edge = edge+1;
			}
		}
	}
	if (CREATE_NEW_TREE_MST==1) {
		tree_g = utreeCreateMinimumSpanningTree(graphcopy,source,g);
	}
	else {
		tree_g = utreeCreateShortestPathTree(graphcopy,source,g);
	}
	ugraphFree(graphcopy);
	g = aggregationBitmapToIntlist(group);
	if (tree_g==NULL) {
		available = 0;
		a->groups[a->numberOfGroups] = 0;
	}
	else {
		found = 0;
		i = (float)utreeTotalCost(tree_g);
		maxEdgeCost = ugraphGetMaxEdgeCost(a->graph);
		while ((i<maxEdgeCost*a->graph->nodeNumber) && (i<=utreeTotalCost(tree_g)*(1+STA2_BOUND)) && (found<NUMBER_OF_CANDIDATES)) {
			t = a->costset[i]->head;
			while ((t!=NULL) && (found<NUMBER_OF_CANDIDATES)) {
				a->numberOfComparisons = a->numberOfComparisons+1;
				maxComparisons = maxComparisons+1;
				if (((~group)|(a->treebitmap[t->value]))==(AGGREGATION_BITMAP)-1) {
					if (bandwidthCheckGroup(a->bandwidth,g,a->trees[t->value],groupRequirement)) { // STA -> STA3
						min = value;
						candidates[found] = t->value;
						found = found+1;
					}
				}
				t = t->next;
			}
			if (found<NUMBER_OF_CANDIDATES) {
				i = i+1;
			}
		}
		//printf("min %.2f\n",min);
		if (maxComparisons>a->maxNumberOfComparisons) {
			a->maxNumberOfComparisons = maxComparisons;
		}
		if (min==AGGREGATION_INFINITY) {
			if (bandwidthAddGroup(a->bandwidth,g,tree_g,groupRequirement)) {
				//printf("bandwidthAddgroup OK\n");
				intlistAddToBack(a->costset[utreeTotalCost(tree_g)],a->numberOfTrees);
				aggregationAddNewTree(a,group);
				//printf("STA new tree of cost %d\n",utreeTotalCost(tree_g)); 
				//utreeDisplay(tree_g);
				//printf("\n");
				//utreeDisplay(tree_g);
				//printf("STA: tree_g\n");
				//printf("\n");
				 
			}
			else {
				available = 0;
				a->groups[a->numberOfGroups] = 0;
				//printf("STA tree_g not good\n");
				//utreeDisplay(tree_g); printf("\n");
			}
		}
		else {
			min_tree = aggregationCandidate(candidates);
#ifdef AGGREGATION_NOTREE
			QERROR("TODO");
#else
			min_tree_candidate = a->trees[min_tree];
			//printf("mintree STA: %d\n",min_tree);
#endif
			
			if (bandwidthAddGroup(a->bandwidth,g,min_tree_candidate,groupRequirement)) {
				aggregationAggregateToTree(a,min_tree);	
				//printf("STA:tree found of cost %d for the group with %d members\n",utreeTotalCost(a->trees[min_tree]),intlistSize(g));
				//utreeDisplay(a->trees[min_tree]);
				//printf("\n");
				//printf("ok pour STA aggregation\n");
			}
			else {
				//printf("no agg\n");
				available = 0;
				a->groups[a->numberOfGroups] = 0;
				//printf("STAmin_tree not good\n");
				//utreeDisplay(a->trees[min_tree]);
				//printf("\n");
			}
		}
	}
	a->numberOfGroups++;
	if (available) {
		a->numberOfAcceptedGroups = a->numberOfAcceptedGroups+1;
	}
	else{
	  //printf("STA: group of size %d refused with tree of cost %d\n",intlistSize(g),utreeTotalCost(tree_g));  
	 
	}
	if (tree_g!=NULL) {
		utreeFree(tree_g);
	}
	intlistFree(g);
	return available;
}

int aggregationAddGroupAQOSM(struct aggregation * a,AGGREGATION_BITMAP group,int groupRequirement){
	int available = 1;
	int i,source,l,k;
	struct utree * tree_g;
	struct utree * t;
	int minTree = -1;
	struct intlist * g;
	int edge;
	struct ugraph * graphcopy;	
	int mincost = AGGREGATION_INFINITY;
	a->groups[a->numberOfGroups] = group;
	a->groupRequirements[a->numberOfGroups] = groupRequirement;
	a->nativeTreeCost[a->numberOfGroups] = 0;
	g = aggregationBitmapToIntlist(group);
	source = g->head->value;
	graphcopy = ugraphCreateFromFile(MAIN_FILE);
	edge = 0;
	for (k=0;k<a->graph->nodeNumber;k++) {
		for (l=0;l<=k;l++) {
			if (ugraphIsEdge(a->graph,k,l)) {
				if (a->bandwidth->residual[edge]<groupRequirement) {
					ugraphRemoveEdges(graphcopy,k,l);
				}
				edge = edge+1;
			}
		}
	}
	if (CREATE_NEW_TREE_MST==1) {
		tree_g = utreeCreateMinimumSpanningTree(graphcopy,source,g);
	}
	else {
		tree_g = utreeCreateShortestPathTree(graphcopy,source,g);
	}
	ugraphFree(graphcopy);
	if(tree_g == NULL){
		available = 0;
		a->groups[a->numberOfGroups] = 0;
	}

	if((tree_g != NULL) && bandwidthCheckGroup(a->bandwidth,g,tree_g,groupRequirement)){
		a->nativeTreeCost[a->numberOfGroups]= utreeTotalCost(tree_g);
		for(i=0;i<a->numberOfTrees;i++){
			t = a->trees[i];
			a->numberOfComparisons = a->numberOfComparisons+1; 
			if ( utreeCoverListOfNodes(t,g) && (bandwidthCheckGroup(a->bandwidth,g,t,groupRequirement)) ) { 
				if (delta_AQOSM(a,g,groupRequirement,i)<=AQOSM_BOUND){
					if(utreeTotalCost(t)<mincost){
						mincost = utreeTotalCost(t);
						minTree = i;
					}
				}
			}
		}
		if (minTree == -1){
			aggregationAddTree(a,group,tree_g);
			//printf("QOS : pas d'agrgation\n");
			bandwidthAddGroup(a->bandwidth,g,tree_g,groupRequirement);
			available = 1;
		}
		else {
			aggregationAggregateToTree(a,minTree);
			//printf("QOS : agrgation\n");
			bandwidthAddGroup(a->bandwidth,g,a->trees[minTree],groupRequirement);
			available = 1;
		}
	}
	else{
		available = 0;
		a->groups[a->numberOfGroups]=0;
	}
	a->numberOfGroups++;
	if(available){	
		//printf("QOS : ");
		//utreeDisplay(a->trees[a->groupToTree[a->numberOfGroups-1]]);
		//printf("\n");
		a->numberOfAcceptedGroups = a->numberOfAcceptedGroups + 1;
	}
	return available;
}

void aggregationAddTree(struct aggregation * a,AGGREGATION_BITMAP group,struct utree * tree_g){
  	if (a==NULL) {
		QERROR("parameters");
	}
	if (a->numberOfTrees+1>a->maxNumberOfTrees) {
		QERROR("too many trees");
	}
#ifdef AGGREGATION_NOTREE	
	a->treecost[a->numberOfTrees] = utreeTotalCost(tree_g);
#else
	a->trees[a->numberOfTrees] = tree_g;
#endif
#ifdef AGGREGATION_NOTREE
	a->treenumberofedges[a->numberOfTrees] = utreeNumberOfEdges(tree_g);
	a->treenumberofspannednodes[a->numberOfTrees] = utreeNumberOfSpannedNodes(tree_g);
	a->treebitmap[a->numberOfTrees] = aggregationTreeBitmap(tree_g,a->numberEdgeRouters);
	utreeFree(tree_g);
#else
	a->treebitmap[a->numberOfTrees] = aggregationTreeBitmap(a->trees[a->numberOfTrees],a->numberEdgeRouters);
#endif
	a->groupToTree[a->numberOfGroups]= a->numberOfTrees;
	a->numberOfAggregatedGroups[a->numberOfTrees] = 1;
	a->numberOfTrees = a->numberOfTrees+1;
}

float delta_AQOSM(struct aggregation * a,struct intlist * group,int groupRequirement,int tree_0){
	float result= 0.0;
	int i = 0;
	int sum_b = 0;
	int sum_bc;
	sum_bc = groupRequirement*a->nativeTreeCost[a->numberOfGroups];
	for(i=0;i<a->numberOfGroups;i++){
		if (a->groups[i]!=0){
			if(a->groupToTree[i]==tree_0){
				sum_b = sum_b + a->groupRequirements[i];
				sum_bc = sum_bc + (a->groupRequirements[i]*a->nativeTreeCost[i]);
			}
		}
	}
	result = utreeTotalCost(a->trees[tree_0])*(sum_b+groupRequirement);
	result = result / sum_bc;
	result = result - 1;
	return result;
}

/*Returns a Candidate among all the candidates in "candidates"*/
int aggregationCandidate(int * candidates){
	int i;
	i = NUMBER_OF_CANDIDATES-1;
	while ((i>=0) && (candidates[i]==-1)) {
		i = i-1;
	}
	if (i<0) {
		QERROR("aggregationCandidate: no candidates");
	}
	return candidates[i];
}


/* adds "group" to "a" without aggregation as in traditional IP multicast*/
int aggregationAddGroupNoAggregation(struct aggregation * a,AGGREGATION_BITMAP group,int groupRequirement) {
	int source;
	int available = 1;
	struct intlist * g;
	struct utree * tree_g;
	g = aggregationBitmapToIntlist(group);
	source = g->head->value;
	
	if (CREATE_NEW_TREE_MST==1) {
		tree_g = utreeCreateMinimumSpanningTree(a->graph,source,g);
	}
	else {
		tree_g = utreeCreateShortestPathTree(a->graph,source,g);
	}
	a->groupRequirements[a->numberOfGroups] = groupRequirement;
	a->nativeTreeCost[a->numberOfGroups] = utreeTotalCost(tree_g);
	if (bandwidthAddGroup(a->bandwidth,g,tree_g,groupRequirement)) {
		a->groups[a->numberOfGroups] = group;
		aggregationAddNewTree(a,group);
		
	}
	else {
	  
		available = 0;
		a->groups[a->numberOfGroups] = 0;
	}
	a->numberOfGroups++;
	if (available) {
		a->numberOfAcceptedGroups = a->numberOfAcceptedGroups+1;
		//printf("NO-  new tree of cost %d\n",utreeTotalCost(tree_g));
	}
	else{
	  //printf("NO-: group with %d members refused with the tree of cost %d\n",intlistSize(g),utreeTotalCost(tree_g));
	} 
	utreeFree(tree_g);
	intlistFree(g);
	return available;
}


/*************************************************************************************************/
/*******************   STA functions : Bitmaps     ***********************************************/
/*************************************************************************************************/



/* transforms the bitmap "bitmap" into a list of members */
struct intlist * aggregationBitmapToIntlist(AGGREGATION_BITMAP bitmap) {
	struct intlist * group;
	int i;
	AGGREGATION_BITMAP mask = 1;
	group = intlistCreate();
	for (i=0;i<8*sizeof(AGGREGATION_BITMAP);i++) {
		if ((bitmap&mask)!=0) {
			intlistAddToBack(group,i);
		}
		mask = 2*mask;
	}
	return group;
}

/* transforms the list of members "group" into a bitmap */
AGGREGATION_BITMAP aggregationIntlistToBitmap(struct intlist * group) {
	AGGREGATION_BITMAP mask = 1;
	AGGREGATION_BITMAP bitmap = 0;
	int i;
	struct intlist_element * member;
	if (group==NULL) {
		QERROR("parameters");
	}
	member = group->head;
	while (member!=NULL) {
		mask = 1;
		for (i=0;i<member->value;i++) {
			mask = 2*mask;
		}
		bitmap = bitmap|mask;
		member = member->next;
	}
	return bitmap;
}

/* returns the bitmap corresponding to the nodes covered by "tree" */
AGGREGATION_BITMAP aggregationTreeBitmap(struct utree * tree,int numberEdgeRouters) {
	int i;
	AGGREGATION_BITMAP mask = 1;
	AGGREGATION_BITMAP bitmap = 0;
	for (i=0;i<numberEdgeRouters;i++) {
		if (utreeCoverNode(tree,i)) {
			bitmap = bitmap|mask;
		}
		mask = 2*mask;
	}
	return bitmap;
}

struct nodeWeightedModel * nodeWeightedModelCreate(int nodeNumber) {
#define LOW_VALUE 20
#define HIGH_VALUE 60
#define HIGH_PERCENTAGE 20
	struct nodeWeightedModel * model;
	struct intlist * highNodes;
	struct intlist_element * next;
	int i;
	if (nodeNumber<0) {
		QERROR("parameters");
	}
	QERROR_MALLOC(model,1,struct nodeWeightedModel);
	model->nodeNumber = nodeNumber;
	QERROR_MALLOC(model->nodeWeight,nodeNumber,int);
	for (i=0;i<nodeNumber;i++) {
		model->nodeWeight[i] = LOW_VALUE;
	}
	highNodes = randomList(HIGH_PERCENTAGE*nodeNumber/100,nodeNumber);
	next = highNodes->head;
	while (next!=NULL) {
		model->nodeWeight[next->value] = HIGH_VALUE;
		next = next->next;
	}
	intlistFree(highNodes);
	return model;
#undef LOW_VALUE
#undef HIGH_VALUE
#undef HIGH_PERCENTAGE
}

void nodeWeightedModelFree(struct nodeWeightedModel * model) {
	if (model==NULL) {
		QERROR("parameters");
	}
	free(model->nodeWeight);
}

struct intlist * nodeWeightedModelNewGroup(struct nodeWeightedModel * model,int groupSize) {
	int i,sum,r,j;
	struct intlist * members;
	if (model==NULL) {
		QERROR("parameters");
	}
	members = intlistCreate();
	for (i=0;i<groupSize;i++) {
		sum = 0;
		for (j=0;j<model->nodeNumber;j++) {
			if (!intlistContain(members,j)) {
				sum = sum+model->nodeWeight[j];
			}
		}
		r = randomGet(sum);
		j = 0;
		while (r>=0) {
			if (!intlistContain(members,j)) {
				r = r-model->nodeWeight[j];
			}
			j = j+1;
		}
		intlistPush(members,j-1);
	}
	return members;
}
