#include <stdio.h>
#include "bandwidth.h"
#include "qerror.h"
#include "random.h"

struct bandwidth * bandwidthCreate(struct ugraph * graph){
  	struct bandwidth * b;
	int i;
	int bandwidth;
	if (graph==NULL) {
		QERROR("parameters");
	}
	QERROR_MALLOC(b,1,struct bandwidth);
	b->graph = graph;
	QERROR_MALLOC(b->capacity,graph->edgeNumber,int);
	QERROR_MALLOC(b->residual,graph->edgeNumber,int);
	for (i=0;i<graph->edgeNumber;i++) {
	  bandwidth = graph->bandwidthCapacity[i];
	  b->capacity[i] = bandwidth;
	  b->residual[i] = b->capacity[i];
	}
	return b;
}

void bandwidthFree(struct bandwidth * b) {
	free(b->capacity);
	free(b->residual);
	free(b);
}
struct bandwidth * bandwidthCopy(struct bandwidth * b){
  struct bandwidth * tmp;
   
  int i;
  if (b==NULL) {
    QERROR("parameters");
  }
  QERROR_MALLOC(tmp,1,struct bandwidth);
  tmp->graph = b->graph;
  QERROR_MALLOC(tmp->capacity,b->graph->edgeNumber,int);
  QERROR_MALLOC(tmp->residual,b->graph->edgeNumber,int);
  for (i=0;i<b->graph->edgeNumber;i++) {
    tmp->capacity[i] = b->capacity[i];
    tmp->residual[i] = b->residual[i];
  }
  return tmp;
  

}


void bandwidthDisplay(struct bandwidth * b) {
	int i;
	int saturated=0;
	int totalUse =0;
	if (b==NULL) {
		QERROR("parameters");
	}			
	for (i=0;i<b->graph->edgeNumber;i++) {
	  totalUse = totalUse + (b->capacity[i]-b->residual[i]);
	  if (b->residual[i]==0){
	    saturated = saturated + 1;
	  }
	  printf("(%.2f ",(100.0*(b->capacity[i]-b->residual[i]))/b->capacity[i]);
	  printf("%d )  - ",b->residual[i]);
	}
	printf("total Use %d and %d saturated edges \n",totalUse,saturated);
	printf("\n");
}

int bandwidthCheckGroup(struct bandwidth * b,struct intlist * group,struct utree * tree,int groupRequirement) {
	int available = 1;
	int i,j,edge;
	if ((b==NULL) || (group==NULL) || (tree==NULL) || (groupRequirement<0)) {
		QERROR("parameters");
	}
	edge = 0;
	for (i=0;i<b->graph->nodeNumber;i++) {
		for (j=0;j<=i;j++) {
			if (ugraphIsEdge(b->graph,i,j)) {
				if (utreeContainEdge(tree,i,j)) {
					if (b->residual[edge]<groupRequirement) {
						available = 0;
					}
				}
				edge = edge+1;
			}
		}
	}
	return available;
}

int bandwidthAddGroup(struct bandwidth * b,struct intlist * group,struct utree * tree,int groupRequirement) {
	int i,j,edge;
	if (bandwidthCheckGroup(b,group,tree,groupRequirement)) {
	  //printf("group accepted check ok\n");
		edge = 0;
		
		for (i=0;i<b->graph->nodeNumber;i++) {
			for (j=0;j<=i;j++) {
				if (ugraphIsEdge(b->graph,i,j)) {
					if (utreeContainEdge(tree,i,j)) {
						b->residual[edge] = b->residual[edge]-groupRequirement;
					}
					edge = edge+1;
				}
			}
		}
		return 1;
	
	}
	else {
	  
		return 0;
	}
}

int bandwidthRemoveGroup(struct bandwidth * b,struct intlist * group,struct utree * tree,int groupRequirement) {
	int i,j,edge;
	if (bandwidthCheckGroup(b,group,tree,groupRequirement)) {
		edge = 0;
		for (i=0;i<b->graph->nodeNumber;i++) {
			for (j=0;j<=i;j++) {
				if (ugraphIsEdge(b->graph,i,j)) {
					if (utreeContainEdge(tree,i,j)) {
						b->residual[edge] = b->residual[edge]+groupRequirement;
					}
					edge = edge+1;
				}
			}
		}
		return 1;
	}
	else {
		return 0;
	}
}
