#include "intlist.h"
#include "qerror.h"
#include <stdio.h>

/* creates a new intintlist */
struct intlist * intlistCreate() {
	struct intlist * l;
	QERROR_MALLOC(l,1,struct intlist);
	l->size = 0;
	l->head = NULL;
	l->tail = NULL;
	return l;
}

/* frees "l" */
void intlistFree(struct intlist * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	while (!intlistIsEmpty(l)) {
		intlistRemoveFront(l);
	}
	free(l);
}

/* makes a copy of "l" */
struct intlist * intlistCopy(struct intlist * l) {
	struct intlist * copy;
	struct intlist_element * element;
	if (l==NULL) {
		QERROR("parameters");
	}
	copy = intlistCreate();
	element = l->head;
	while (element!=NULL) {
		intlistAddToBack(copy,element->value);
		element = element->next;
	}
	return copy;
}

/* returns the number of elements of "l" */
int intlistLength(struct intlist * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return intlistSize(l);
}

/* returns the number of elements of "l" */
int intlistSize(struct intlist * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return l->size;
}

/* checks if "l" is empty */
int intlistIsEmpty(struct intlist * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	if (l->size==0) {
		return 1;
	}
	else {
		return 0;
	}
}

/* returns the first element of "l" */
struct intlist_element * intlistFirst(struct intlist * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return l->head;
}

/* returns the last element of "l" */
struct intlist_element * intlistLast(struct intlist * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return l->tail;
}

/* returns the "position"-th element of "l" */
struct intlist_element * intlistGet(struct intlist * l,int position) {
	struct intlist_element * element;
	int i;
	if ((l==NULL) || (position<0) || (position>=intlistSize(l))) {
		QERROR("parameters");
	}
	i = 0;
	element = l->head;
	for (i=0;i<position;i++) {
		element = element->next;
	}
	return element;
}

/* returns the successor of "element" in "l" */
struct intlist_element * intlistSucc(struct intlist * l,struct intlist_element * element) {
	if ((l==NULL) || (element==NULL)) {
		QERROR("parameters");
	}
	return element->next;
}

/* returns the predecessor of "element" in "l" */
struct intlist_element * intlistPred(struct intlist * l,struct intlist_element * element) {
	if ((l==NULL) || (element==NULL)) {
		QERROR("parameters");
	}
	return element->previous;
}

/* returns the cyclic successor of "element" in "l" (the successor of "element" in "l" if defined, the first element of "l" otherwise") */
struct intlist_element * intlistCyclicSucc(struct intlist * l,struct intlist_element * element) {
	if ((l==NULL) || (element==NULL)) {
		QERROR("parameters");
	}
	if (element->next==NULL) {
		return l->head;
	}
	else {
		return element->next;
	}
}

/* returns the cyclic predecessor of "element" in "l" (the predecessor of "element" in "l" if defined, the last element of "l" otherwise") */
struct intlist_element * intlistCyclicPred(struct intlist * l,struct intlist_element * element) {
	if ((l==NULL) || (element==NULL)) {
		QERROR("parameters");
	}
	if (element->previous==NULL) {
		return l->tail;
	}
	else {
		return element->previous;
	}
}

/* removes the "position"-th element of "l" */
void intlistRemove(struct intlist * l,int position) {
	struct intlist_element * element;
	if ((l==NULL) || (position<0) || (position>=intlistSize(l))) {
		QERROR("parameters");
	}
	element = intlistGet(l,position);
	intlistDelete(l,element);
}

/* returns the first value of "l" */
int intlistFront(struct intlist * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	if (intlistIsEmpty(l)) {
		QERROR("empty intlist");
	}
	return l->head->value;
}

/* returns the first value of "l" */
int intlistHead(struct intlist * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return intlistFront(l);
}

/* returns the last value of "l" */
int intlistBack(struct intlist * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	if (intlistIsEmpty(l)) {
		QERROR("empty intlist");
	}
	return l->tail->value;
}

/* returns the last value of "l" */
int intlistTail(struct intlist * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return intlistBack(l);
}

/* adds a new element with value "value" to the front of "l" */
struct intlist_element * intlistAddToFront(struct intlist * l,int value) {
	struct intlist_element * element;
	if (l==NULL) {
		QERROR("parameters");
	}
	QERROR_MALLOC(element,1,struct intlist_element);
	element->value = value;
	if (l->head!=NULL) {
		l->head->previous = element;
	}
	element->previous = NULL;
	element->next = l->head;
	l->head = element;
	if (l->tail==NULL) {
		l->tail = element;
	}
	l->size = l->size+1;
	return element;
}

/* adds a new element with value "value" to the back of "l" */
struct intlist_element * intlistAddToBack(struct intlist * l,int value) {
	struct intlist_element * element;
	if (l==NULL) {
		QERROR("parameters");
	}
	QERROR_MALLOC(element,1,struct intlist_element);
	element->value = value;
	if (l->tail!=NULL) {
		l->tail->next = element;
	}
	element->previous = l->tail;
	element->next = NULL;
	l->tail = element;
	if (l->head==NULL) {
		l->head = element;
	}
	l->size = l->size+1;
	return element;
}

/* adds a new element with value "value" to the front of "l" */
struct intlist_element * intlistPush(struct intlist * l,int value) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return intlistAddToFront(l,value);
}

/* adds a new element with value "value" to the back of "l" */
struct intlist_element * intlistAppend(struct intlist * l,int value) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return intlistAddToBack(l,value);
}

/* adds a new element with value "value" after "position" in "l" */
struct intlist_element * intlistInsertAfter(struct intlist * l,struct intlist_element * position,int value) {
	struct intlist_element * element;
	if ((l==NULL) || (position==NULL)) {
		QERROR("parameters");
	}
	if (position->next==NULL) {
		return intlistAddToBack(l,value);
	}
	else {
		QERROR_MALLOC(element,1,struct intlist_element);
		element->value = value;
		element->previous = position;
		element->next = position->next;
		position->next = element;
		element->next->previous = element;
		l->size = l->size+1;
		return element;
	}
}

/* adds a new element with value "value" before "position" in "l" */
struct intlist_element * intlistInsertBefore(struct intlist * l,struct intlist_element * position,int value) {
	struct intlist_element * element;
	if ((l==NULL) || (position==NULL)) {
		QERROR("parameters");
	}
	if (position->previous==NULL) {
		return intlistAddToFront(l,value);
	}
	else {
		QERROR_MALLOC(element,1,struct intlist_element);
		element->value = value;
		element->previous = position->previous;
		element->next = position;
		position->previous = element;
		element->previous->next = element;
		l->size = l->size+1;
		return element;
	}
}

/* removes the first element of "l" */
void intlistRemoveFront(struct intlist * l) {
	struct intlist_element * element;
	if (l==NULL) {
		QERROR("parameters");
	}
	if (intlistIsEmpty(l)) {
		QERROR("empty intlist");
	}
	element = l->head;
	l->head = element->next;
	if (l->head==NULL) {
		l->tail = NULL;
	}
	else {
		l->head->previous = NULL;
	}
	free(element);
	l->size = l->size-1;
}

/* removes the last element of "l" */
void intlistRemoveBack(struct intlist * l) {
	struct intlist_element * element;
	if (l==NULL) {
		QERROR("parameters");
	}
	if (intlistIsEmpty(l)) {
		QERROR("empty intlist");
	}
	element = l->tail;
	l->tail = element->previous;
	if (l->tail==NULL) {
		l->head = NULL;
	}
	else {
		l->tail->next = NULL;
	}
	free(element);
	l->size = l->size-1;
}

/* returns the first element of "l" */
int intlistPop(struct intlist * l) {
	int value;
	if (l==NULL) {
		QERROR("parameters");
	}
	if (intlistIsEmpty(l)) {
		QERROR("empty intlist");
	}
	value = l->head->value;
	intlistRemoveFront(l);
	return value;
}

/* removes the "position"-th element of "l" */
void intlistDeletePosition(struct intlist * l,int position) {
	struct intlist_element * element;
	if ((l==NULL) || (position<0) || (position>=intlistSize(l))) {
		QERROR("parameters");
	}
	element = intlistGet(l,position);
	intlistDelete(l,element);
	//ON A ENLEVE LE FREE(ELEMENT)
}

/* removes all the first element having value "value" */
void intlistDeleteValue(struct intlist * l,int value) {
	struct intlist_element * element;
	int found = 0;
	if (l==NULL) {
		QERROR("parameters");
	}
	element = l->head;
	while ((element!=NULL) && (!found)) {
		if (element->value==value) {
			found = 1;
		}
		else {
			element = element->next;
		}
	}
	intlistDelete(l,element);
}

/* removes the element "element" of "l" */
void intlistDelete(struct intlist * l,struct intlist_element * element) {
	if ((l==NULL) || (element==NULL)) {
		QERROR("parameters");
	}
	if (element->previous==NULL) {
		intlistRemoveFront(l);
	}
	else {
		if (element->next==NULL) {
			intlistRemoveBack(l);
		}
		else {
			element->previous->next = element->next;
			element->next->previous = element->previous;
			free(element);
			l->size = l->size-1;
		}
	}
}

/* moves "element" to the front of "l" */
void intlistMoveToFront(struct intlist * l,struct intlist_element * element) {
	int value;
	value = element->value;
	intlistDelete(l,element);
	intlistAddToFront(l,value);
}

/* moves "element" to the back of "l" */
void intlistMoveToBack(struct intlist * l,struct intlist_element * element) {
	int value;
	value = element->value;
	intlistDelete(l,element);
	intlistAddToBack(l,value);
}

/* assigns the value "newValue" to the element "element" in "l" */
void intlistAssign(struct intlist * l,struct intlist_element * element,int newValue) {
	if ((l==NULL) || (element==NULL)) {
		QERROR("parameters");
	}
	element->value = newValue;
}

/* puts the elements of "addedList" at the end of "l", and clears "addedList" */
void intlistConcAfter(struct intlist * l,struct intlist * addedList) {
	if ((l==NULL) || (addedList==NULL)) {
		QERROR("parameters");
	}
	if (intlistIsEmpty(l)) {
		l->head = addedList->head;
		l->tail = addedList->tail;
	}
	else {
		l->tail->next = addedList->head;
		addedList->head->previous = l->tail;
		l->tail = addedList->tail;
	}
	l->size = l->size+addedList->size;
	addedList->size = 0;
	addedList->head = NULL;
	addedList->tail = NULL;
}

/* puts the elements of "addedList" at the beginning of "l", and clears "addedList" */
void intlistConcBefore(struct intlist * l,struct intlist * addedList) {
	if ((l==NULL) || (addedList==NULL)) {
		QERROR("parameters");
	}
	if (intlistIsEmpty(l)) {
		l->head = addedList->head;
		l->tail = addedList->tail;
	}
	else {
		addedList->tail->next = l->head;
		l->head->previous = addedList->tail;
		l->head = addedList->head;
	}
	l->size = l->size+addedList->size;
	addedList->size = 0;
	addedList->head = NULL;
	addedList->tail = NULL;
}

/* reverses "l" */
void intlistReverse(struct intlist * l) {
	struct intlist_element * element;
	struct intlist_element * tmp;
	if (l==NULL) {
		QERROR("parameters");
	}
	element = l->head;
	while (element!=NULL) {
		tmp = element->next;
		element->next = element->previous;
		element->previous = tmp;
		element = tmp;
	}
	tmp = l->tail;
	l->tail = l->head;
	l->head = tmp;
}

/* clears "l" */
void intlistClear(struct intlist * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	while (!intlistIsEmpty(l)) {
		intlistRemoveFront(l);
	}
}

/* sorts "l" by the minimum selection algorithm */
void intlistSort(struct intlist * l) {
	struct intlist * sortedList;
	struct intlist_element * element;
	struct intlist_element * minimum_element;
	sortedList = intlistCreate();
	while (!intlistIsEmpty(l)) {
		minimum_element = l->head;
		element = l->head->next;
		while (element!=NULL) {
			if (element->value<minimum_element->value) {
				minimum_element = element;
			}
			element = element->next;
		}
		intlistAddToBack(sortedList,minimum_element->value);
		intlistDelete(l,minimum_element);
	}
	l->head = sortedList->head;
	l->tail = sortedList->tail;
	l->size = sortedList->size;
	free(sortedList);
}

/* searchs for "value" in "l" */
struct intlist_element * intlistSearch(struct intlist * l,int value) {
	struct intlist_element * element;
	int found;
	found = 0;
	element = l->head;
	while ((!found) && (element!=NULL)) {
		if (element->value==value) {
			found = 1;
		}
		else {
			element = element->next;
		}
	}
	if (found) {
		return element;
	}
	else {
		return NULL;
	}
}

/* searchs for "value" in "l" */
struct intlist_element * intlistFind(struct intlist * l,int value) {
	return intlistSearch(l,value);
}

/* displays "l" */
void intlistDisplay(struct intlist * l) {
	struct intlist_element * element;
	if (l==NULL) {
		QERROR("parameters");
	}
	element = l->head;
	printf("( ");
	while (element!=NULL) {
		printf("%d ",element->value);
		element = element->next;
	}
	printf(")");
}

/* checks if "l" contains "value" */
int intlistContain(struct intlist * l,int value) {
	if (intlistSearch(l,value)==NULL) {
		return 0;
	}
	else {
		return 1;
	}
}

/* checks if "l1" and "l2" contains the same members in the same order */
int intlistEquals(struct intlist * l1,struct intlist * l2) {
	struct intlist_element * e1,* e2;
	int equal = 1;
	if ((l1==NULL) || (l2==NULL)) {
		QERROR("parameters");
	}
	e1 = l1->head;
	e2 = l2->head;
	while ((equal==1) && (e1!=NULL) && (e2!=NULL)) {
		if (e1->value!=e2->value) {
			equal = 0;
		}
		e1 = e1->next;
		e2 = e2->next;
	}
	if (((e1==NULL) && (e2!=NULL)) || ((e1!=NULL) && (e2==NULL))) {
		equal = 0;
	}
	return equal;
}

/* checks if "l1" is included in "l2" */
int intlistIsContainedIn(struct intlist * l1,struct intlist * l2) {
	int contained;
	struct intlist_element * e1;
	if ((l1==NULL) || (l2==NULL)) {
		QERROR("parameters");
	}
	contained = 1;
	e1 = l1->head;
	while ((contained==1) && (e1!=NULL)) {
		if (!intlistContain(l2,e1->value)) {
			contained = 0;
		}
		e1 = e1->next;
	}
	return contained;
}

/* compares "l1" and "l2" and returns the lexicographical order between "l1" and "l2" ("0" if "l1=l2", "-1" if "l1<l2" and "1" if "l1>l2" */
int intlistCompare(struct intlist * l1,struct intlist * l2) {
	struct intlist_element * e1,* e2;
	int order = 0;
	if ((l1==NULL) || (l2==NULL)) {
		QERROR("parameters");
	}
	e1 = l1->head;
	e2 = l2->head;
	while ((order==0) && (e1!=NULL) && (e2!=NULL)) {
		if (e1->value<e2->value) {
			order = -1;
		}
		else {
			if (e1->value>e2->value) {
				order = 1;
			}
			else {
				e1 = e1->next;
				e2 = e2->next;
			}
		}
	}
	if (order==0) {
		if (e1==NULL) {
			if (e2==NULL) {
				order = 0;
			}
			else {
				order = -1;
			}
		}
		else {
			order = 1;
		}
	}
	return order;
}

/* inserts "value" in the sorted list "l" */
void intlistSortedInsert(struct intlist * l,int value) {
	struct intlist_element * element;
	if (l==NULL) {
		QERROR("parameters");
	}
	element = l->head;
	while ((element!=NULL) && (element->value<value)) {
		element = element->next;
	}
	if (element==NULL) {
		intlistAddToBack(l,value);
	}
	else {
		intlistInsertBefore(l,element,value);
	}
}

/* returns the number of elements that are in "l1" but not in "l2" */
int intlistDifference(struct intlist * l1,struct intlist * l2) {
	int number = 0;
	struct intlist_element * element;
	element = l1->head;
	while (element!=NULL) {
		if (!intlistContain(l2,element->value)) {
			number = number+1;
		}
		element = element->next;
	}
	return number;
}
