#ifndef INTLIST_H
#define INTLIST_H

struct intlist_element {
	int value;
	struct intlist_element * previous;
	struct intlist_element * next;
};

struct intlist {
	int size;
	struct intlist_element * head;
	struct intlist_element * tail;
};

/* creation and destruction operations */

struct intlist * intlistCreate();
void intlistFree(struct intlist * l);

struct intlist * intlistCopy(struct intlist * l);

/* access operations */

int intlistLength(struct intlist * l);
int intlistSize(struct intlist * l);
int intlistIsEmpty(struct intlist * l);

struct intlist_element * intlistFirst(struct intlist * l);
struct intlist_element * intlistLast(struct intlist * l);

struct intlist_element * intlistGet(struct intlist * l,int position);

struct intlist_element * intlistSucc(struct intlist * l,struct intlist_element * element);
struct intlist_element * intlistPred(struct intlist * l,struct intlist_element * element);
struct intlist_element * intlistCyclicSucc(struct intlist * l,struct intlist_element * element);
struct intlist_element * intlistCyclicPred(struct intlist * l,struct intlist_element * element);

void intlistRemove(struct intlist * l,int position);

int intlistFront(struct intlist * l);
int intlistHead(struct intlist * l);
int intlistBack(struct intlist * l);
int intlistTail(struct intlist * l);

int intlistContain(struct intlist * l,int value);

/* update operations */

struct intlist_element * intlistAddToFront(struct intlist * l,int value);
struct intlist_element * intlistAddToBack(struct intlist * l,int value);
struct intlist_element * intlistPush(struct intlist * l,int value);
struct intlist_element * intlistAppend(struct intlist * l,int value);
struct intlist_element * intlistInsertAfter(struct intlist * l,struct intlist_element * position,int value);
struct intlist_element * intlistInsertBefore(struct intlist * l,struct intlist_element * position,int value);

void intlistRemoveFront(struct intlist * l);
void intlistRemoveBack(struct intlist * l);

int intlistPop(struct intlist * l);

void intlistDeletePosition(struct intlist * l,int position);
void intlistDelete(struct intlist * l,struct intlist_element * element);
void intlistDeleteValue(struct intlist * l,int value);

void intlistMoveToFront(struct intlist * l,struct intlist_element * element);
void intlistMoveToBack(struct intlist * l,struct intlist_element * element);

void intlistAssign(struct intlist * l,struct intlist_element * element,int newValue);

void intlistConcAfter(struct intlist * l,struct intlist * addedList);
void intlistConcBefore(struct intlist * l,struct intlist * addedList);

void intlistReverse(struct intlist * l);

void intlistClear(struct intlist * l);

int intlistDifference(struct intlist * l1,struct intlist * l2);

/* sorting and searching */

void intlistSort(struct intlist * l);

void intlistSortedInsert(struct intlist * l,int value);

struct intlist_element * intlistSearch(struct intlist * l,int value);
struct intlist_element * intlistFind(struct intlist * l,int value);

void intlistDisplay(struct intlist * l);

int intlistEquals(struct intlist * l1,struct intlist * l2);

int intlistIsContainedIn(struct intlist * l1,struct intlist * l2);

int intlistCompare(struct intlist * l1,struct intlist * l2);

#endif
