#include "list.h"
#include "qerror.h"
#include <stdio.h>

/* creates a new list */
struct list * listCreate(void (* listFreeElement)(void *),void * (listCopyElement)(void *),void (* listDisplayElement)(void *)) {
	struct list * l;
	QERROR_MALLOC(l,1,struct list);
	l->size = 0;
	l->head = NULL;
	l->tail = NULL;
	l->listFreeElement = listFreeElement;
	l->listCopyElement = listCopyElement;
	l->listDisplayElement = listDisplayElement;
	return l;
}

/* frees "l" */
void listFree(struct list * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	while (!listIsEmpty(l)) {
		listRemoveFront(l);
	}
	free(l);
}

/* makes a copy of "l" */
struct list * listCopy(struct list * l) {
	struct list * copy;
	struct list_element * element;
	if (l==NULL) {
		QERROR("parameters");
	}
	copy = listCreate(l->listFreeElement,l->listCopyElement,l->listDisplayElement);
	element = l->head;
	while (element!=NULL) {
		listAddToBack(copy,l->listCopyElement(element->value));
		element = element->next;
	}
	return copy;
}

/* returns the number of elements of "l" */
int listLength(struct list * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return listSize(l);
}

/* returns the number of elements of "l" */
int listSize(struct list * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return l->size;
}

/* checks if "l" is empty */
int listIsEmpty(struct list * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	if (l->size==0) {
		return 1;
	}
	else {
		return 0;
	}
}

/* returns the first element of "l" */
struct list_element * listFirst(struct list * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return l->head;
}

/* returns the last element of "l" */
struct list_element * listLast(struct list * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return l->tail;
}

/* returns the "position"-th element of "l" */
struct list_element * listGet(struct list * l,int position) {
	struct list_element * element;
	int i;
	if ((l==NULL) || (position<0) || (position>=listSize(l))) {
		QERROR("parameters");
	}
	i = 0;
	element = l->head;
	for (i=0;i<position;i++) {
		element = element->next;
	}
	return element;
}

/* returns the successor of "element" in "l" */
struct list_element * listSucc(struct list * l,struct list_element * element) {
	if ((l==NULL) || (element==NULL)) {
		QERROR("parameters");
	}
	return element->next;
}

/* returns the predecessor of "element" in "l" */
struct list_element * listPred(struct list * l,struct list_element * element) {
	if ((l==NULL) || (element==NULL)) {
		QERROR("parameters");
	}
	return element->previous;
}

/* returns the cyclic successor of "element" in "l" (the successor of "element" in "l" if defined, the first element of "l" otherwise") */
struct list_element * listCyclicSucc(struct list * l,struct list_element * element) {
	if ((l==NULL) || (element==NULL)) {
		QERROR("parameters");
	}
	if (element->next==NULL) {
		return l->head;
	}
	else {
		return element->next;
	}
}

/* returns the cyclic predecessor of "element" in "l" (the predecessor of "element" in "l" if defined, the last element of "l" otherwise") */
struct list_element * listCyclicPred(struct list * l,struct list_element * element) {
	if ((l==NULL) || (element==NULL)) {
		QERROR("parameters");
	}
	if (element->previous==NULL) {
		return l->tail;
	}
	else {
		return element->previous;
	}
}

/* removes the "position"-th element of "l" */
void listRemove(struct list * l,int position) {
	struct list_element * element;
	if ((l==NULL) || (position<0) || (position>=listSize(l))) {
		QERROR("parameters");
	}
	element = listGet(l,position);
	listDelete(l,element);
}

/* returns the first value of "l" */
void * listFront(struct list * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	if (listIsEmpty(l)) {
		QERROR("empty list");
	}
	return l->head->value;
}

/* returns the first value of "l" */
void * listHead(struct list * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return listFront(l);
}

/* returns the last value of "l" */
void * listBack(struct list * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	if (listIsEmpty(l)) {
		QERROR("empty list");
	}
	return l->tail->value;
}

/* returns the last value of "l" */
void * listTail(struct list * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return listBack(l);
}

/* adds a new element with value "value" (a copy if "listCopyElement" is not NULL) to the front of "l" */
struct list_element * listAddToFront(struct list * l,void * value) {
	struct list_element * element;
	if (l==NULL) {
		QERROR("parameters");
	}
	QERROR_MALLOC(element,1,struct list_element);
	if (l->listCopyElement==NULL) {
		element->value = value;
	}
	else {
		element->value = l->listCopyElement(value);
	}
	if (l->head!=NULL) {
		l->head->previous = element;
	}
	element->previous = NULL;
	element->next = l->head;
	l->head = element;
	if (l->tail==NULL) {
		l->tail = element;
	}
	l->size = l->size+1;
	return element;
}

/* adds a new element with value "value" (a copy if "listCopyElement" is not NULL) to the back of "l" */
struct list_element * listAddToBack(struct list * l,void * value) {
	struct list_element * element;
	if (l==NULL) {
		QERROR("parameters");
	}
	QERROR_MALLOC(element,1,struct list_element);
	if (l->listCopyElement==NULL) {
		element->value = value;
	}
	else {
		element->value = l->listCopyElement(value);
	}
	if (l->tail!=NULL) {
		l->tail->next = element;
	}
	element->previous = l->tail;
	element->next = NULL;
	l->tail = element;
	if (l->head==NULL) {
		l->head = element;
	}
	l->size = l->size+1;
	return element;
}

/* adds a new element with value "value" (a copy if "listCopyElement" is not NULL) to the front of "l" */
struct list_element * listPush(struct list * l,void * value) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return listAddToFront(l,value);
}

/* adds a new element with value "value" (a copy if "listCopyElement" is not NULL) to the back of "l" */
struct list_element * listAppend(struct list * l,void * value) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return listAddToBack(l,value);
}

/* adds a new element with value "value" (a copy if "listCopyElement" is not NULL) after "position" in "l" */
struct list_element * listInsertAfter(struct list * l,struct list_element * position,void * value) {
	struct list_element * element;
	if ((l==NULL) || (position==NULL)) {
		QERROR("parameters");
	}
	if (position->next==NULL) {
		return listAddToBack(l,value);
	}
	else {
		QERROR_MALLOC(element,1,struct list_element);
		if (l->listCopyElement==NULL) {
			element->value = value;
		}
		else {
			element->value = l->listCopyElement(value);
		}
		element->previous = position;
		element->next = position->next;
		position->next = element;
		element->next->previous = element;
		l->size = l->size+1;
		return element;
	}
}

/* adds a new element with value "value" (a copy if "listCopyElement" is not NULL) before "position" in "l" */
struct list_element * listInsertBefore(struct list * l,struct list_element * position,void * value) {
	struct list_element * element;
	if ((l==NULL) || (position==NULL)) {
		QERROR("parameters");
	}
	if (position->previous==NULL) {
		return listAddToFront(l,value);
	}
	else {
		QERROR_MALLOC(element,1,struct list_element);
		if (l->listCopyElement==NULL) {
			element->value = value;
		}
		else {
			element->value = l->listCopyElement(value);
		}
		element->previous = position->previous;
		element->next = position;
		position->previous = element;
		element->previous->next = element;
		l->size = l->size+1;
		return element;
	}
}

/* removes the first element of "l" */
void listRemoveFront(struct list * l) {
	struct list_element * element;
	if (l==NULL) {
		QERROR("parameters");
	}
	if (listIsEmpty(l)) {
		QERROR("empty list");
	}
	element = l->head;
	l->head = element->next;
	if (l->head==NULL) {
		l->tail = NULL;
	}
	else {
		l->head->previous = NULL;
	}
	if (l->listFreeElement!=NULL) {
		l->listFreeElement(element->value);
	}
	l->size = l->size-1;
	free(element);
}

/* removes the last element of "l" */
void listRemoveBack(struct list * l) {
	struct list_element * element;
	if (l==NULL) {
		QERROR("parameters");
	}
	if (listIsEmpty(l)) {
		QERROR("empty list");
	}
	element = l->tail;
	l->tail = element->previous;
	if (l->tail==NULL) {
		l->head = NULL;
	}
	else {
		l->tail->next = NULL;
	}
	if (l->listFreeElement!=NULL) {
		l->listFreeElement(element->value);
	}
	l->size = l->size-1;
	free(element);
}

/* returns the first element of "l" (does not free it) */
void * listPop(struct list * l) {
	void * value;
	if (l==NULL) {
		QERROR("parameters");
	}
	if (listIsEmpty(l)) {
		QERROR("empty list");
	}
	if (l->listCopyElement==NULL) {
		value = l->head->value;
	}
	else {
		value = l->listCopyElement(l->head->value);
	}
	listRemoveFront(l);
	return value;
}

/* removes the "position"-th element of "l" */
void listDeletePosition(struct list * l,int position) {
	struct list_element * element;
	if ((l==NULL) || (position<0) || (position>=listSize(l))) {
		QERROR("parameters");
	}
	element = listGet(l,position);
	listDelete(l,element);
}

/* removes the element "element" of "l" */
void listDelete(struct list * l,struct list_element * element) {
	if ((l==NULL) || (element==NULL)) {
		QERROR("parameters");
	}
	if (element->previous==NULL) {
		listRemoveFront(l);
	}
	else {
		if (element->next==NULL) {
			listRemoveBack(l);
		}
		else {
			element->previous->next = element->next;
			element->next->previous = element->previous;
			if (l->listFreeElement!=NULL) {
				l->listFreeElement(element->value);
			}
			free(element);
			l->size = l->size-1;
		}
	}
}

/* moves "element" to the front of "l" */
void listMoveToFront(struct list * l,struct list_element * element) {
	void * value;
	if (l->listCopyElement==NULL) {
		value = element->value;
	}
	else {
		value = l->listCopyElement(value);
	}
	listDelete(l,element);
	listAddToFront(l,value);
}

/* moves "element" to the back of "l" */
void listMoveToBack(struct list * l,struct list_element * element) {
	void * value;
	if (l->listCopyElement==NULL) {
		value = element->value;
	}
	else {
		value = l->listCopyElement(value);
	}
	listDelete(l,element);
	listAddToBack(l,value);
}

/* assigns the value "newValue" to the element "element" in "l" */
void listAssign(struct list * l,struct list_element * element,void * newValue) {
	if ((l==NULL) || (element==NULL)) {
		QERROR("parameters");
	}
	element->value = newValue;
}

/* puts the elements of "addedList" at the end of "l", and clears "addedList" */
void listConcAfter(struct list * l,struct list * addedList) {
	if ((l==NULL) || (addedList==NULL)) {
		QERROR("parameters");
	}
	if (listIsEmpty(l)) {
		l->head = addedList->head;
		l->tail = addedList->tail;
	}
	else {
		l->tail->next = addedList->head;
		addedList->head->previous = l->tail;
		l->tail = addedList->tail;
	}
	l->size = l->size+addedList->size;
	addedList->size = 0;
	addedList->head = NULL;
	addedList->tail = NULL;
}

/* puts the elements of "addedList" at the beginning of "l", and clears "addedList" */
void listConcBefore(struct list * l,struct list * addedList) {
	if ((l==NULL) || (addedList==NULL)) {
		QERROR("parameters");
	}
	if (listIsEmpty(l)) {
		l->head = addedList->head;
		l->tail = addedList->tail;
	}
	else {
		addedList->tail->next = l->head;
		l->head->previous = addedList->tail;
		l->head = addedList->head;
	}
	l->size = l->size+addedList->size;
	addedList->size = 0;
	addedList->head = NULL;
	addedList->tail = NULL;
}

/* reverses "l" */
void listReverse(struct list * l) {
	struct list_element * element;
	struct list_element * tmp;
	if (l==NULL) {
		QERROR("parameters");
	}
	element = l->head;
	while (element!=NULL) {
		tmp = element->next;
		element->next = element->previous;
		element->previous = tmp;
		element = tmp;
	}
	tmp = l->tail;
	l->tail = l->head;
	l->head = tmp;
}

/* clears "l" */
void listClear(struct list * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	while (!listIsEmpty(l)) {
		listRemoveFront(l);
	}
}

/* sorts "l" by the minimum selection algorithm, using the function "compare" ("compare(a,b)" equals "0" if "a"="b", is lower than "0" if "a"<"b" and is greater than "0" if "a">"b") */
void listSort(struct list * l,int (* compare)(void *,void *)) {
	struct list * sortedList;
	struct list_element * element;
	struct list_element * minimum_element;
	sortedList = listCreate(l->listFreeElement,l->listCopyElement,l->listDisplayElement);
	while (!listIsEmpty(l)) {
		minimum_element = l->head;
		element = l->head->next;
		while (element!=NULL) {
			if (compare(element->value,minimum_element->value)<0) {
				minimum_element = element;
			}
			element = element->next;
		}
		listAddToBack(sortedList,minimum_element->value);
		listDelete(l,minimum_element);
	}
	l->head = sortedList->head;
	l->tail = sortedList->tail;
	l->size = sortedList->size;
	free(sortedList);
}

/* searchs for "value" in "l", using the function "compare"  */
struct list_element * listSearch(struct list * l,void * value,int (* compare)(void *,void *)) {
	struct list_element * element;
	int found;
	found = 0;
	element = l->head;
	while ((!found) && (element!=NULL)) {
		if (compare(element->value,value)==0) {
			found = 1;
		}
		else {
			element = element->next;
		}
	}
	if (found) {
		return element;
	}
	else {
		return NULL;
	}
}

/* searchs for "value" in "l" */
struct list_element * listSearchStandard(struct list * l,void * value) {
	return listSearch(l,value,listFindCompareEqual);
}

/* displays "l" */
void listDisplay(struct list * l) {
	struct list_element * element;
	if (l==NULL) {
		QERROR("parameters");
	}
	element = l->head;
	printf("( ");
	while (element!=NULL) {
		l->listDisplayElement(element->value);
		printf(" ");
		element = element->next;
	}
	printf(")");
}

/* checks if "a" and "b" are the same pointers */
int listFindCompareEqual(void * a,void * b) {
	if (a==b) {
		return 0;
	}
	else {
		if (a<b) {
			return -1;
		}
		else {
			return 1;
		}
	}
}
