#ifndef LIST_H
#define LIST_H

struct list_element {
	void * value;
	struct list_element * previous;
	struct list_element * next;
};

struct list {
	int size;
	struct list_element * head;
	struct list_element * tail;
	void (* listFreeElement)(void *);
	void * (* listCopyElement)(void *);
	void (* listDisplayElement)(void *);
};

/* creation and destruction operations */

struct list * listCreate(void (* listFreeElement)(void *),void * (listCopyElement)(void *),void (* listDisplayElement)(void *));
void listFree(struct list * l);

struct list * listCopy(struct list * l);

/* access operations */

int listLength(struct list * l);
int listSize(struct list * l);
int listIsEmpty(struct list * l);

struct list_element * listFirst(struct list * l);
struct list_element * listLast(struct list * l);

struct list_element * listGet(struct list * l,int position);

struct list_element * listSucc(struct list * l,struct list_element * element);
struct list_element * listPred(struct list * l,struct list_element * element);
struct list_element * listCyclicSucc(struct list * l,struct list_element * element);
struct list_element * listCyclicPred(struct list * l,struct list_element * element);

void listRemove(struct list * l,int position);

void * listFront(struct list * l);
void * listHead(struct list * l);
void * listBack(struct list * l);
void * listTail(struct list * l);

/* update operations */

struct list_element * listAddToFront(struct list * l,void * value);
struct list_element * listAddToBack(struct list * l,void * value);
struct list_element * listPush(struct list * l,void * value);
struct list_element * listAppend(struct list * l,void * value);
struct list_element * listInsertAfter(struct list * l,struct list_element * position,void * value);
struct list_element * listInsertBefore(struct list * l,struct list_element * position,void * value);

void listRemoveFront(struct list * l);
void listRemoveBack(struct list * l);

void * listPop(struct list * l);

void listDeletePosition(struct list * l,int position);
void listDelete(struct list * l,struct list_element * element);

void listMoveToFront(struct list * l,struct list_element * element);
void listMoveToBack(struct list * l,struct list_element * element);

void listAssign(struct list * l,struct list_element * element,void * newValue);

void listConcAfter(struct list * l,struct list * addedList);
void listConcBefore(struct list * l,struct list * addedList);

void listReverse(struct list * l);

void listClear(struct list * l);

/* sorting and searching */

void listSort(struct list * l,int (* compare)(void *,void *));

struct list_element * listSearch(struct list * l,void * value,int (* compare)(void *,void *));
struct list_element * listSearchStandard(struct list * l,void * value);

void listDisplay(struct list * l);

/* internal */

int listFindCompareEqual(void * a,void * b);

#endif
