#include "rbtree.h"
#include "qerror.h"
#include "intlist.h"
#include "stack.h"
#include <stdio.h>
#include <stdlib.h>

/* creates a new red-black tree */
struct rbtree * rbtreeCreate(void (* elementFree)(void *),void * (* elementCopy)(void *),void (* elementDisplay)(void *),int (* elementCompare)(void *,void *)) {
	struct rbtree * rbt;
	if (elementCompare==NULL) {
		QERROR("parameters");
	}
	QERROR_MALLOC(rbt,1,struct rbtree);
	rbt->root = NULL;
	rbt->size = 0;
	rbt->elementFree = elementFree;
	rbt->elementCopy = elementCopy;
	rbt->elementDisplay = elementDisplay;
	rbt->elementCompare = elementCompare;
	return rbt;
}

/* frees "rbt" */
void rbtreeFree(struct rbtree * rbt) {
	// ...
}

/* displays "rbt" */
void rbtreeDisplay(struct rbtree * rbt) {
	struct stack * stack;
	struct rbtree_element * element;
	if (rbt==NULL) {
		QERROR("parameters");
	}
	stack = stackCreate(NULL,NULL,NULL);
	printf("[ ");
	if (rbt->root!=NULL) {
		stackPush(stack,rbt->root);
	}
	while (!stackIsEmpty(stack)) {
		element = (struct rbtree_element *)stackPop(stack);
		printf("(");
		printf("node=%p ",element);
		if (element->color==RBTREE_BLACK) {
			printf("color=black ");
		}
		else {
			if (element->color==RBTREE_RED) {
				printf("color=red ");
			}
			else {
				printf("color=? ");
			}
		}
		printf("value=");
		rbt->elementDisplay(element->value);
		if (element->left!=NULL) {
			stackPush(stack,element->left);
		}
		if (element->right!=NULL) {
			stackPush(stack,element->right);
		}
		printf(" ");
		printf("left=%p ",element->left);
		printf("right=%p",element->right);
		printf(") ");
	}
	printf("]");
	stackFree(stack);
}

/* copies "rbt" */
struct rbtree * rbtreeCopy(struct rbtree * rbt) {
	// ...
	return rbt;
}

/* inserts "element" in "rbt" */
void rbtreeInsert(struct rbtree * rbt,int id,void * element,int * numberOfComparisons) {
	struct rbtree_element * x,* y;
	if (rbt==NULL) {
		QERROR("parameters");
	}
	QERROR_MALLOC(x,1,struct rbtree_element);
	if (rbt->elementCopy==NULL) {
		x->value = element;
	}
	else {
		x->value = rbt->elementCopy(element);
	}
	rbtreeNormalInsert(rbt,x,numberOfComparisons);
	x->id = id;
	x->color = RBTREE_RED;
	while ((x!=rbt->root) && (x->parent->color==RBTREE_RED)) {
		if (x->parent==x->parent->parent->left) {
			y = x->parent->parent->right;
			if ((y==NULL) || (y->color==RBTREE_RED)) { // !!! ajout de "(y==NULL)"
				x->parent->color = RBTREE_BLACK;
				if (y!=NULL) y->color = RBTREE_BLACK; // !!! ajout de "if (y!=NULL)"
				x->parent->parent->color = RBTREE_RED;
				x = x->parent->parent;
			}
			else {
				if (x==x->parent->right) {
					x = x->parent;
					rbtreeLeftRotation(rbt,x);
				}
				x->parent->color = RBTREE_BLACK;
				x->parent->parent->color = RBTREE_RED;
				rbtreeRightRotation(rbt,x->parent->parent);
			}
		}
		else {
			y = x->parent->parent->left;
			if ((y==NULL) || (y->color==RBTREE_RED)) { // !!! ajout de "(y==NULL)"
				x->parent->color = RBTREE_BLACK;
				if (y!=NULL) y->color = RBTREE_BLACK; // !!! ajout de "if (y!=NULL)"
				x->parent->parent->color = RBTREE_RED;
				x = x->parent->parent;
			}
			else {
				if (x==x->parent->left) {
					x = x->parent;
					rbtreeRightRotation(rbt,x);
				}
				x->parent->color = RBTREE_BLACK;
				x->parent->parent->color = RBTREE_RED;
				rbtreeLeftRotation(rbt,x->parent->parent);
			}
		}
	}
	rbt->root->color = RBTREE_BLACK;
}

/* searches for "element" in "rbt" and return the id of the element if it is found, "-1" elsewhere */
int rbtreeSearch(struct rbtree * rbt,void * element,int * numberOfComparisons) {
	struct rbtree_element * x;
	int compareResult;
	int found = 0;
	if (rbt==NULL) {
		QERROR("parameters");
	}
	x = rbt->root;
	while ((!found) && (x!=NULL)) {
		compareResult = rbt->elementCompare(element,x->value);
		if (numberOfComparisons!=NULL) {
			*numberOfComparisons = *numberOfComparisons+1;
		}
		if (compareResult==0) {
			found = 1;
		}
		else {
			if (compareResult<0) {
				x = x->left;
			}
			else {
				x = x->right;
			}
		}
	}
	if (found) {
		return x->id;
	}
	else {
		return -1;
	}
}

/* delete from "rbt" element "element" */
void rbtreeDelete(struct rbtree * rbt,void * element,int * numberOfComparisons) {
	// ...
}

/* performs a left rotation of "rbt" on 'x" */
void rbtreeLeftRotation(struct rbtree * rbt,struct rbtree_element * x) {
	struct rbtree_element * y;
	if ((rbt==NULL) || (x==NULL) || (x->right==NULL)) {
		QERROR("parameters");
	}
	y = x->right;
	x->right = y->left;
	if (y->left!=NULL) {
		y->left->parent = x;
	}
	y->parent = x->parent;
	if (x->parent==NULL) {
		rbt->root = y;
	}
	else {
		if (x==x->parent->left) {
			x->parent->left = y;
		}
		else {
			x->parent->right = y;
		}
	}
	y->left = x;
	x->parent = y;
}

/* performs a right rotation of "rbt" on 'x" */
void rbtreeRightRotation(struct rbtree * rbt,struct rbtree_element * x) {
	struct rbtree_element * y;
	if ((rbt==NULL) || (x==NULL) || (x->left==NULL)) {
		QERROR("parameters");
	}
	y = x->left;
	x->left = y->right;
	if (y->right!=NULL) {
		y->right->parent = x;
	}
	y->parent = x->parent;
	if (x->parent==NULL) {
		rbt->root = y;
	}
	else {
		if (x==x->parent->right) {
			x->parent->right = y;
		}
		else {
			x->parent->left = y;
		}
	}
	y->right = x;
	x->parent = y;
}

/* inserts "x" in "rbt" as if "rbt" was an unbalanced binary search tree */
void rbtreeNormalInsert(struct rbtree * rbt,struct rbtree_element * x,int * numberOfComparisons) {
	struct rbtree_element * parent,* current;
	int compareResult;
	if ((rbt==NULL) || (x==NULL)) {
		QERROR("parameters");
	}
	x->color = RBTREE_UNDEFINED;
	x->left = NULL;
	x->right = NULL;
	parent = NULL;
	current = rbt->root;
	while (current!=NULL) {
		compareResult = rbt->elementCompare(x->value,current->value);
		if (numberOfComparisons!=NULL) {
			*numberOfComparisons = *numberOfComparisons+1;
		}
		if (compareResult<0) {
			parent = current;
			current = current->left;
		}
		else {
			parent = current;
			current = current->right;
		}
	}
	x->parent = parent;
	if (x->parent==NULL) {
		rbt->root = x;
	}
	else {
		if (compareResult<0) {
			parent->left = x;
		}
		else {
			parent->right = x;
		}
	}
	rbt->size = rbt->size+1;
}

/* fixup "rbt" on element "x", after a deletion */
void rbtreeDeleteFixup(struct rbtree * rbt,struct rbtree_element * x,int * numberOfComparisons) {
	// ...
}
