#ifndef RBTREE_H
#define RBTREE_H

struct rbtree {
	struct rbtree_element * root;
	int size;
	void (* elementFree)(void * element);
	void * (* elementCopy)(void * element);
	void (* elementDisplay)(void * element);
	int (* elementCompare)(void * element1,void * element2);
};

#define RBTREE_UNDEFINED 0
#define RBTREE_BLACK 1
#define RBTREE_RED 2

struct rbtree_element {
	int color;
	int id; // the number of the group
	void * value;
	struct rbtree_element * left;
	struct rbtree_element * right;
	struct rbtree_element * parent;
};

struct rbtree * rbtreeCreate(void (* elementFree)(void *),void * (* elementCopy)(void *),void (* elementDisplay)(void *),int (* elementCompare)(void *,void *));
struct rbtree * rbtreeCopy(struct rbtree * rbt);

void rbtreeFree(struct rbtree * rbt);

void rbtreeDisplay(struct rbtree * rbt);

void rbtreeInsert(struct rbtree * rbt,int id,void * element,int * numberOfComparisons);

int rbtreeSearch(struct rbtree * rbt,void * element,int * numberOfComparisons);

void rbtreeDelete(struct rbtree * rbt,void * element,int * numberOfComparisons);

void rbtreeNormalInsert(struct rbtree * rbt,struct rbtree_element * x,int * numberOfComparisons);
void rbtreeLeftRotation(struct rbtree * rbt,struct rbtree_element * x);
void rbtreeRightRotation(struct rbtree * rbt,struct rbtree_element * x);
void rbtreeDeleteFixup(struct rbtree * rbt,struct rbtree_element * x,int * numberOfComparisons);

#endif
