#include "stack.h"
#include "qerror.h"

/* creates an empty stack */
struct stack * stackCreate(void (* stackFreeElement)(void *),void * (* stackCopyElement)(void *),void (* stackDisplayElement)(void *)) {
	struct stack * s;
	QERROR_MALLOC(s,1,struct stack);
	s->l = listCreate(stackFreeElement,stackCopyElement,stackDisplayElement);
	return s;
}

/* frees "s" */
void stackFree(struct stack * s) {
	if (s==NULL) {
		QERROR("parameters");
	}
	listFree(s->l);
	free(s);
}

/* copies "s" into a new stack */
struct stack * stackCopy(struct stack * s) {
	struct stack * copy;
	if (s==NULL) {
		QERROR("parameters");
	}
	QERROR_MALLOC(copy,1,struct stack);
	copy->l = listCopy(s->l);
	return copy;
}

/* returns the top of "s" */
void * stackTop(struct stack * s) {
	if (s==NULL) {
		QERROR("parameters");
	}
	return listFront(s->l);
}

/* pushes "value" to "s" */
void stackPush(struct stack * s,void * value) {
	if (s==NULL) {
		QERROR("parameters");
	}
	listPush(s->l,value);
}

/* popes "s" (does not free the value) */
void * stackPop(struct stack * s) {
	if (s==NULL) {
		QERROR("parameters");
	}
	return listPop(s->l);
}

/* returns the size of "s" */
int stackSize(struct stack * s) {
	if (s==NULL) {
		QERROR("parameters");
	}
	return listSize(s->l);
}

/* checks if "s" is empty */
int stackIsEmpty(struct stack * s) {
	if (s==NULL) {
		QERROR("parameters");
	}
	return listIsEmpty(s->l);
}

/* clears "s" */
void stackClear(struct stack * s) {
	if (s==NULL) {
		QERROR("parameters");
	}
	listClear(s->l);
}

/* displays "s" */
void stackDisplay(struct stack * s) {
	if (s==NULL) {
		QERROR("parameters");
	}
	listDisplay(s->l);
}
