#include "ugraph.h"
#include "qerror.h"
#include "random.h"
#include "intstack.h"
#include <stdio.h>
#include <math.h>

/* creates a new ugraph */
struct ugraph * ugraphCreate(int nodeNumber,void (* ugraphFreeElement)(void *),void * (* ugraphCopyElement)(void *),void (* ugraphDisplayElement)(void *)) {
	struct ugraph * g;
	int i;
	QERROR_MALLOC(g,1,struct ugraph);
	g->nodeNumber = nodeNumber;
	g->edgeNumber = 0;
	QERROR_MALLOC(g->nodes,nodeNumber,struct alist *);
	QERROR_MALLOC(g->bandwidthCapacity,nodeNumber*nodeNumber,int);
	QERROR_MALLOC(g->nodeCapacities,nodeNumber,struct nodeInfo *);
	for (i=0;i<nodeNumber;i++) {
		QERROR_MALLOC(g->nodeCapacities[i],1,struct nodeInfo);
	}
	for (i=0;i<nodeNumber;i++) {
		g->nodes[i] = alistCreate(ugraphFreeElement,ugraphCopyElement,ugraphDisplayElement);
	}
	g->ugraphFreeElement = ugraphFreeElement;
	g->ugraphCopyElement = ugraphCopyElement;
	g->ugraphDisplayElement = ugraphDisplayElement;
	return g;
}
/*

struct ugraph * ugraphCreateFromFile(char * filename) {
	struct ugraph * g;
	int nodeNumber,edgeNumber,i;
	int sourceNode,destinationNode,edgeCost;
	int * cost;
	FILE * f;
	f = fopen(filename,"r");
	if (f==NULL) {
		QERROR("file does not exist");
	}
	fscanf(f,"%d %d\n",&nodeNumber,&edgeNumber);
	g = ugraphCreate(nodeNumber,ugraphFreeInt,ugraphCopyInt,ugraphDisplayInt);
	
	for (i=0;i<edgeNumber;i++) {
		fscanf(f,"%d %d %d\n",&sourceNode,&destinationNode,&edgeCost);
		QERROR_MALLOC(cost,1,int);
		*cost = edgeCost;
		ugraphAddEdge(g,sourceNode,destinationNode,cost);
		free(cost);
	}
	fclose(f);
	return g;
}
*/

/* creates a new ugraph from the file "filename" */
struct ugraph * ugraphCreateFromFile(char * filename) {
	struct ugraph * g;
	int nodeNumber,edgeNumber,i;
	int sourceNode,destinationNode,edgeCost,bandwidth;
	int * cost;
	int realValues = 0;
	FILE * f;

	f = fopen(filename,"r");
	if (f==NULL) {
		QERROR("file does not exist");
	}
	
	fscanf(f,"%d %d %d\n",&nodeNumber,&edgeNumber,&realValues);
	g = ugraphCreate(nodeNumber,ugraphFreeInt,ugraphCopyInt,ugraphDisplayInt);
	
	for (i=0;i<edgeNumber;i++) {
	  if (realValues == 1){
	    fscanf(f,"%d %d %d %d\n",&sourceNode,&destinationNode,&edgeCost,&bandwidth);
	    QERROR_MALLOC(cost,1,int);
	    *cost = edgeCost;
	    ugraphAddEdge(g,sourceNode,destinationNode,cost);
	    g->bandwidthCapacity[i] = bandwidth;
	    free(cost);

	  }
	  else{
	    fscanf(f,"%d %d %d\n",&sourceNode,&destinationNode,&edgeCost);
	    QERROR_MALLOC(cost,1,int);
	    *cost = edgeCost;
	    ugraphAddEdge(g,sourceNode,destinationNode,cost);
	    g->bandwidthCapacity[i]=randomGet(20000)+100000;
	    free(cost);
	    
	  }
	  
	}
	fclose(f);
	return g;
}
/* creates an ugraph using Doar (alpha=0.25 and beta=0.2)*/
struct ugraph * ugraphCreateDoar(int nodeNumber,float alpha,float beta) {
	struct ugraph * g;
	int * cost;
	int i,j,edgecost;
	int * x,* y;
	if ((nodeNumber<=0) || (alpha<=0) || (beta<=0)) {
		QERROR("parameters");
	}
	QERROR_MALLOC(x,nodeNumber,int);
	QERROR_MALLOC(y,nodeNumber,int);
	g = NULL;
	do {
		if (g!=NULL) {
			ugraphFree(g);
		}
		for (i=0;i<nodeNumber;i++) {
			x[i] = randomGet(100);
			y[i] = randomGet(100);
		}
		g = ugraphCreate(nodeNumber,ugraphFreeInt,ugraphCopyInt,ugraphDisplayInt);
		for (i=0;i<nodeNumber;i++) {
			for (j=0;j<i;j++) {
				float value = (float)((double)rand()/((double)RAND_MAX+1)*1);
				if (value<beta*exp((-sqrt((x[i]-x[j])*(x[i]-x[j])+(y[i]-y[j])*(y[i]-y[j])))/(alpha*100*sqrt(2)))) {
					edgecost = (int)(sqrt((x[i]-x[j])*(x[i]-x[j])+(y[i]-y[j])*(y[i]-y[j])))+1;
					QERROR_MALLOC(cost,1,int);
					*cost = edgecost;
					ugraphAddEdge(g,i,j,cost);
					free(cost);
				}
			}
		}
	} while (!ugraphIsConnected(g));
	free(x);
	free(y);
	return g;
}


/* stores the graph into "filename" */
void ugraphStoreToFile(struct ugraph * g,char * filename) {
	FILE * f;
	int i,j;
	f = fopen(filename,"w");
	if (f==NULL) {
		QERROR("error opening file");
	}
	fprintf(f,"%d %d\n",g->nodeNumber,g->edgeNumber);
	for (i=0;i<g->nodeNumber;i++) {
		for (j=0;j<i;j++) {
			if (ugraphIsEdge(g,i,j)) {
				fprintf(f,"%d %d %d\n",i,j,ugraphIntEdgeCost(ugraphGetEdgeCost(g,i,j)));
			}
		}
	}
	fclose(f);
}



/* creates an ugraph using Doar, with constant degree e (alpha=0.25, beta=0.2 and k=25)*/
struct ugraph * ugraphCreateDoarConstantDegree(int nodeNumber,float alpha,float beta,float k,float e) {
	struct ugraph * g;
	int * cost;
	int i,j,edgecost;
	int * x,* y;
	if ((nodeNumber<=0) || (alpha<=0) || (beta<=0) || (k<=0) || (e<=0)) {
		QERROR("parameters");
	}
	QERROR_MALLOC(x,nodeNumber,int);
	QERROR_MALLOC(y,nodeNumber,int);
	g = NULL;
	do {
		if (g!=NULL) {
			ugraphFree(g);
		}
		for (i=0;i<nodeNumber;i++) {
			x[i] = randomGet(100);
			y[i] = randomGet(100);
		}
		g = ugraphCreate(nodeNumber,ugraphFreeInt,ugraphCopyInt,ugraphDisplayInt);
		for (i=0;i<nodeNumber;i++) {
			for (j=0;j<i;j++) {
				float value = (float)((double)rand()/((double)RAND_MAX+1)*1);
				if (value<(k*e/nodeNumber)*beta*exp((-sqrt((x[i]-x[j])*(x[i]-x[j])+(y[i]-y[j])*(y[i]-y[j])))/(alpha*100*sqrt(2)))) {
					edgecost = (int)(sqrt((x[i]-x[j])*(x[i]-x[j])+(y[i]-y[j])*(y[i]-y[j])))+1;
					QERROR_MALLOC(cost,1,int);
					*cost = edgecost;
					ugraphAddEdge(g,i,j,cost);
					free(cost);
				}
			}
		}
	} while (!ugraphIsConnected(g));
	free(x);
	free(y);
	return g;
}

/* frees "g" */
void ugraphFree(struct ugraph * g) {
	int i;
	if (g==NULL) {
		QERROR("parameters");
	}
	for (i=0;i<g->nodeNumber;i++){
		alistFree(g->nodes[i]);
		free(g->nodeCapacities[i]);
	}
	free(g->nodes);
	free(g->bandwidthCapacity);
	free(g->nodeCapacities);
	free(g);
}

/* adds the (undirected) edge between "source" and "destination" (and from "destination" to "source") with cost "cost" (a copy) in "g" */
void ugraphAddEdge(struct ugraph * g,int source,int destination,void * cost) {
	if ((g==NULL) || (source<0) || (source>=g->nodeNumber) || (destination<0) || (destination>=g->nodeNumber)) {
		QERROR("parameters");
	}
	g->edgeNumber = g->edgeNumber+1;
	alistAdd(g->nodes[source],destination,cost);
	alistAdd(g->nodes[destination],source,cost);
}

/* checks if "g" is connected */
int ugraphIsConnected(struct ugraph * g) {
	int * visited;
	int connected,i;
	struct intlist_element * element;
	struct intstack * stack;
	if (g==NULL) {
		QERROR("parameters");
	}
	connected = 1;
	stack = intstackCreate();
	QERROR_MALLOC(visited,g->nodeNumber,int);
	for (i=0;i<g->nodeNumber;i++) {
		visited[i] = 0;
	}
	intstackPush(stack,0);
	visited[0] = 1;
	while (!intstackIsEmpty(stack)) {
		i = intstackPop(stack);
		element = g->nodes[i]->keys->head;
		while (element!=NULL) {
			if (visited[element->value]==0) {
				visited[element->value] = 1;
				intstackPush(stack,element->value);
			}
			element = element->next;
		}
	}
	intstackFree(stack);
	i = 0;
	while ((connected) && (i<g->nodeNumber)) {
		if (visited[i]==0) {
			connected = 0;
		}
		else {
			i++;
		}
	}
	free(visited);
	return connected;
}

/* returns the cost of an edge between "source" and "destination" in "g" */
void * ugraphGetEdgeCost(struct ugraph * g,int source,int destination) {
	if ((g==NULL) || (source<0) || (source>=g->nodeNumber) || (destination<0) || (destination>=g->nodeNumber)) {
		QERROR("parameters");
	}
	return alistGet(g->nodes[source],destination);
}

/* checks if "g" contains an edge between "source" and "destination" */
int ugraphIsEdge(struct ugraph * g,int source,int destination) {
	if ((g==NULL) || (source<0) || (source>=g->nodeNumber) || (destination<0) || (destination>=g->nodeNumber)) {
		QERROR("parameters");
	}
	if (alistContain(g->nodes[source],destination)) {
		return 1;
	}
	else {
		return 0;
	}
}

/* removes all the edges from "source" to "destination" in "g" */
void ugraphRemoveEdges(struct ugraph * g,int source,int destination) {
	if ((g==NULL) || (source<0) || (source>=g->nodeNumber) || (destination<0) || (destination>=g->nodeNumber)) {
		QERROR("parameters");
	}
	g->edgeNumber = g->edgeNumber-alistNumber(g->nodes[source],destination);
	alistRemoveAll(g->nodes[source],destination);
	alistRemoveAll(g->nodes[destination],source);
}

/* displays "g" */
void ugraphDisplay(struct ugraph * g) {
	int i;
	struct intlist_element * next;
	printf("[ ");
	for (i=0;i<g->nodeNumber;i++) {
		next = g->nodes[i]->keys->head;
		while (next!=NULL) {
			if (i<=next->value) {
				printf("(%d,%d) ",i,next->value);
			}
			next = next->next;
		}
	}
	printf("]");
}


/* displays "g" and indicates the cost of each edge */
void ugraphDisplayWithCost(struct ugraph * g) {
	int i;
	struct intlist_element * next;
	printf("[ ");
	for (i=0;i<g->nodeNumber;i++) {
		next = g->nodes[i]->keys->head;
		while (next!=NULL) {
			if (i<=next->value) {
				printf("(%d,%d)",i,next->value);
				printf("<");
				g->ugraphDisplayElement(ugraphGetEdgeCost(g,i,next->value));
				printf("> ");
			}
			next = next->next;
		}
	}
	printf("]");
}

/* returns the degree of node "node" in "g" */
int ugraphGetNodeDegree(struct ugraph * g,int node) {
	if ((g==NULL) || (node<0) || (node>=g->nodeNumber)) {
		QERROR("parameters");
	}
	return alistSize(g->nodes[node]);
}

/* runs the Dijkstra algorithm from "source" on "g", computing the "distances" and the "predecessors" arrays */
void ugraphDijkstra(struct ugraph * g,int source,int * predecessors,int * distances) {
	int i;
	for (i=0;i<g->nodeNumber;i++) {
		distances[i] = UGRAPH_INFINITY;
	}
	distances[source] = 0;
	if (predecessors!=NULL) {
		for (i=0;i<g->nodeNumber;i++) {
			predecessors[i] = -1;
		}
	}
	if (predecessors!=NULL) {
		predecessors[source] = source;
	}
	ugraphDijkstraNoInit(g,predecessors,distances);
}

/* runs the Dijkstra algorithm on "g", computing the "distances" and the "predecessors" array, based on the initialized "distances" array */
void ugraphDijkstraNoInit(struct ugraph * g,int * predecessors,int * distances) {
	int i,min;
	char * visited;
	struct intlist_element * keys_successors;
	struct list_element * value_successors;
	if ((g==NULL) || (distances==NULL)) {
		QERROR("parameters");
		return;
	}
	QERROR_MALLOC(visited,g->nodeNumber,char);
	for (i=0;i<g->nodeNumber;i++) {
		visited[i] = 0;
	}
	do {
		min = -1;
		for (i=0;i<g->nodeNumber;i++) {
			if (visited[i]==0) {
				if ((min==-1) || (distances[i]<distances[min])) {
					min = i;
				}
			}
		}
		if (min!=-1) {
			visited[min] = 1;
			keys_successors = g->nodes[min]->keys->head;
			value_successors = g->nodes[min]->values->head;
			while (keys_successors!=NULL) {
				if (distances[min]+ugraphIntEdgeCost(value_successors->value)<distances[keys_successors->value]) {
					distances[keys_successors->value] = distances[min]+ugraphIntEdgeCost(value_successors->value);
					if (predecessors!=NULL) {
						predecessors[keys_successors->value] = min;
					}
				}
				keys_successors = keys_successors->next;
				value_successors = value_successors->next;
			}
		}
	} while (min!=-1);
	free(visited);
}

/* generates an intlist having "memberPercentage" percentage of members in "g" */
struct intlist * ugraphGenerateMembers(struct ugraph * g,int memberPercentage) {
	int i;
	int random;
	struct intlist * members;
	int * member;
	if (g==NULL) {
		QERROR("parameters");
	}
	QERROR_MALLOC(member,g->nodeNumber,int);
	members = intlistCreate();
	for (i=0;i<g->nodeNumber;i++) {
		member[i] = 0;
	}
	for (i=0;i<g->nodeNumber*memberPercentage/100;i++) {
		random = randomGet(g->nodeNumber);
		while (member[random]==1) {
			random = (random+1)%g->nodeNumber;
		}
		member[random] = 1;
		intlistPush(members,random);
	}
	return members;
}

/* return the number of edges of "g" */
int ugraphGetNumberOfEdges(struct ugraph * g) {
	if (g==NULL) {
		QERROR("parameters");
	}
	return g->edgeNumber;
}

/* displays the int cost of "edge" */
void ugraphDisplayInt(void * edge) {
	int * intedge;
	if (edge==NULL) {
		printf("?");
	}
	else {
		intedge = (int *)edge;
		printf("%d",*intedge);
	}
}

/* frees an int "edge" */
void ugraphFreeInt(void * edge) {
	free(edge);
}

/* copies an int "edge */
void * ugraphCopyInt(void * edge) {
	int * previousEdge;
	int * newEdge;
	if (edge==NULL) {
		return NULL;
	}
	previousEdge = (int *)edge;
	QERROR_MALLOC(newEdge,1,int);
	*newEdge = *previousEdge;
	return newEdge;
}

/* copies "g" */
struct ugraph * ugraphCopy(struct ugraph * g) {
	struct ugraph * new;
	int i;
	new = ugraphCreate(g->nodeNumber,g->ugraphFreeElement,g->ugraphCopyElement,g->ugraphDisplayElement);
	for (i=0;i<g->nodeNumber;i++) {
	  alistFree(new->nodes[i]);
	  new->nodes[i] = alistCopy(g->nodes[i]);
	}
	for (i=0;i<g->nodeNumber;i++) {
		QERROR_MALLOC(new->nodeCapacities[i],1,struct nodeInfo);
		// struct nodeInfo should not contain any pointers
		*(new->nodeCapacities[i]) = *(g->nodeCapacities[i]);
	}
	return new;
}

/* returns the metrical closure of nodes of "nodes" in "g" */
struct ugraph * ugraphMetricalClosure(struct ugraph * g,struct intlist * nodes) {
	struct ugraph * metricalClosure;
	struct intlist_element * tmp;
	int * distances;
	int * cost;
	int i,j,size;
	size = intlistSize(nodes);
	if ((g==NULL) || (nodes==NULL)) {
		QERROR("parameters");
	}
	QERROR_MALLOC(distances,g->nodeNumber,int);
	metricalClosure = ugraphCreate(size,g->ugraphFreeElement,g->ugraphCopyElement,g->ugraphDisplayElement);
	for (i=0;i<size-1;i++) {
		ugraphDijkstra(g,i,NULL,distances);
		tmp = nodes->head;
		for (j=0;j<size;j++) {
			if (j>=i+1) {
				QERROR_MALLOC(cost,1,int);
				*cost = distances[tmp->value];
				ugraphAddEdge(metricalClosure,i,j,(void *)cost);
				free(cost);
			}
			tmp = tmp->next;
		}
	}
	free(distances);
	return metricalClosure;
}

/* returns the int cost of "edge" */
int ugraphIntEdgeCost(void * edge) {
	int * cost;
	if (edge==NULL) {
		return 1;
	}
	else {
		cost = (int *)edge;
		return *cost;
	}
}

/* sets a cost of "1" for each link */
void ugraphSetCostOnePerLink(struct ugraph * g) {
	int i;
	struct list_element * element;
	if (g==NULL) {
		QERROR("parameters");
	}
	for (i=0;i<g->nodeNumber;i++) {
		element = g->nodes[i]->values->head;
		while (element!=NULL) {
			free(element->value);
			element->value = NULL;
			element = element->next;
		}
	}
}

/* returns the maximum cost of an edge of "g" */
int ugraphGetMaxEdgeCost(struct ugraph * g) {
	int max;
	int i,j;
	void * edgeCost;
	if (g==NULL) {
		QERROR("parameters");
	}
	max = 0;
	for (i=0;i<g->nodeNumber;i++) {
		for (j=0;j<g->nodeNumber;j++) {
			if (ugraphIsEdge(g,i,j)) {
				edgeCost = ugraphGetEdgeCost(g,i,j);
				if (max<ugraphIntEdgeCost(edgeCost)) {
					max = ugraphIntEdgeCost(edgeCost);
				}
			}
		}
	}
	return max;
}
