#ifndef UGRAPH_H
#define UGRAPH_H

#include "alist.h"

struct nodeInfo {
};

struct ugraph {
	int nodeNumber;
	int edgeNumber;
  int * bandwidthCapacity;
	struct alist ** nodes;
	struct nodeInfo ** nodeCapacities;
	void (* ugraphFreeElement)(void *);
	void * (* ugraphCopyElement)(void *);
	void (* ugraphDisplayElement)(void *);
};

#define UGRAPH_INFINITY 100000000

struct ugraph * ugraphCreate(int nodeNumber,void (* ugraphFreeElement)(void *),void * (* ugraphCopyElement)(void *),void (* ugraphDisplayElement)(void *));
struct ugraph * ugraphCreateFromFile(char * filename);
void ugraphFree(struct ugraph * g);
struct ugraph * ugraphCopy(struct ugraph * g);
struct ugraph * ugraphCreateDoar(int nodeNumber,float alpha,float beta);
struct ugraph * ugraphCreateDoarConstantDegree(int nodeNumber,float alpha,float beta,float k,float e);

struct intlist * ugraphGenerateMembers(struct ugraph * g,int memberPercentage);
void ugraphStoreToFile(struct ugraph * g,char * filename);
 
int ugraphGetNumberOfEdges(struct ugraph * g);
int ugraphGetNodeDegree(struct ugraph * g,int node);
int ugraphIsEdge(struct ugraph * g,int source,int destination);
void * ugraphGetEdgeCost(struct ugraph * g,int source,int destination);
int ugraphGetMaxEdgeCost(struct ugraph * g);

void ugraphAddEdge(struct ugraph * g,int source,int destination,void * cost);

void ugraphRemoveEdges(struct ugraph * g,int source,int destination);

int ugraphIsConnected(struct ugraph * g);
void ugraphDijkstra(struct ugraph * g,int source,int * predecessors,int * distances);
void ugraphDijkstraNoInit(struct ugraph * g,int * predecessors,int * distances);

void ugraphDisplay(struct ugraph * g);
void ugraphDisplayWithCost(struct ugraph * g);

struct ugraph * ugraphMetricalClosure(struct ugraph * g,struct intlist * nodes);

void ugraphFreeInt(void * edge);
void * ugraphCopyInt(void * edge);
void ugraphDisplayInt(void * edge);

int ugraphIntEdgeCost(void * edge);

void ugraphSetCostOnePerLink(struct ugraph * g);

#endif
