#ifndef UTREE_H
#define UTREE_H

#include "intlist.h"
#include "ugraph.h"

struct utree {
	int source;
	struct ugraph * tree;
};

struct utree * utreeCreate(struct ugraph * support,int source);
struct utree * utreeCreateShortestPathTree(struct ugraph * support,int source,struct intlist * members);
struct utree * utreeCreateMinimumSpanningTree(struct ugraph * support, int source, struct intlist * members);
struct utree * utreeCreateSteinerTree(struct ugraph * support, int source, struct intlist * members);
struct utree * utreeCreateSteinerTree1Member(struct ugraph * support, int source, struct intlist * members);
struct utree * utreeCreateSteinerTree2Members(struct ugraph * support, int source, struct intlist * members);
struct utree * utreeCreateSteinerTree3Members(struct ugraph * support, int source, struct intlist * members);

void utreeFree(struct utree * t);
void utreeDisplay(struct utree * t);

struct utree * utreeCopy(struct utree * t);

void utreeAddEdge(struct utree * t,int node1,int node2,void * cost);

int utreeCoverNode(struct utree * t,int node);
int utreeCoverListOfNodes(struct utree * t,struct intlist * nodes);

int utreeContainEdge(struct utree * t,int node1,int node2);

void utreeDijkstra(struct utree * t,struct ugraph * support,struct intlist * members,int * predecessors,int * distances);

int utreeGetClosestMember(struct utree * t,struct intlist * members,int * distances);
void utreeAddNode(struct utree * t,struct ugraph * support,int node,int * predecessors);

int utreeTotalCost(struct utree * t);

int utreeNumberOfSpannedNodes(struct utree * t);
int utreeNumberOfBranchingNodes(struct utree * t);
int utreeNumberOfEdges(struct utree * t);
int utreeNumberOfLeaves(struct utree * t);

int utreeParent(struct utree * t,int node);

struct intlist * utreeBreadthFirstSearch(struct utree * t);
struct intlist * utreeSubtree(struct utree * t,int node);
void utreeContractInGraph(struct utree * t,struct ugraph * g,int node);

#endif
